#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Address Space."""

import re

import ucdp as u
from pytest import fixture, raises

import ucdp_addr as ua


@fixture
def addrmap():
    """Alias to one."""
    addrmap = ua.AddrMap()
    for idx, aname in enumerate(("uart", "spi", "owi")):
        addrspace = ua.Addrspace(name=aname, width=32, depth=8, baseaddr=8 * 4 * idx)
        addrmap.add(addrspace)
        word = addrspace.add_word("ctrl")
        word.add_field("ena", u.EnaType(), "RW")
        word.add_field("mode", u.UintType(4, default=1), "RW")
        word.add_field("speed", u.UintType(8, default=3), "RW")
        word = addrspace.add_word("stat")
        word.add_field("bsy", u.BitType(), "RO", offset=9)
        word = addrspace.add_word("rc")
        word.add_field("bit", u.BitType(), "RC")
        word = addrspace.add_word("wl")
        word.add_field("bit", u.BitType(), "WL")
    return addrmap


def test_int(addrmap):
    """Integer."""
    item = 0x20

    wordinfo = ua.WordInfo.create(addrmap, item, 5)
    assert str(wordinfo) == "0x20 [0x20 1x32 (4 bytes)] SINGLE: 5"
    assert str(tuple(wordinfo.iter())) == "((Hex('0x20'), 5),)"

    wordinfo = ua.WordInfo.create(addrmap, item, (5, 6))
    assert str(wordinfo) == "0x20 [0x20 2x32 (8 bytes)] BURST: (5, 6)"
    assert str(tuple(wordinfo.iter())) == "((Hex('0x20'), 5), (Hex('0x24'), 6))"

    wordinfo = ua.WordInfo.create(addrmap, item, ((8, 5), (16, 6)))
    assert str(wordinfo) == "0x20 [0x20 5x32 (20 bytes)] SCAT: ((8, 5), (16, 6))"
    assert str(tuple(wordinfo.iter())) == "((Hex('0x28'), 5), (Hex('0x30'), 6))"

    wordinfo = ua.WordInfo.create(addrmap, item, 5, mask=0xF, offset=8)
    assert str(wordinfo) == "0x28 [0x28 1x32 (4 bytes)] mask=0xF SINGLE: 5"
    assert str(tuple(wordinfo.iter())) == "((Hex('0x28'), 5),)"


def test_addrrange(addrmap):
    """Address Range."""
    item = ua.AddrRange(baseaddr=0x20, size=64)

    wordinfo = ua.WordInfo.create(addrmap, item, 5)
    assert str(wordinfo) == "0x20 [0x20 1x32 (4 bytes)] SINGLE: 5"
    assert str(tuple(wordinfo.iter())) == "((Hex('0x20'), 5),)"

    wordinfo = ua.WordInfo.create(addrmap, item, (5, 6))
    assert str(wordinfo) == "0x20 [0x20 2x32 (8 bytes)] BURST: (5, 6)"
    assert str(tuple(wordinfo.iter())) == "((Hex('0x20'), 5), (Hex('0x24'), 6))"
    assert str(tuple(wordinfo.addrs())) == "(Hex('0x20'), Hex('0x24'))"

    wordinfo = ua.WordInfo.create(addrmap, item, ((8, 5), (16, 6)))
    assert str(wordinfo) == "0x20 [0x20 5x32 (20 bytes)] SCAT: ((8, 5), (16, 6))"
    assert str(tuple(wordinfo.iter())) == "((Hex('0x28'), 5), (Hex('0x30'), 6))"
    assert str(tuple(wordinfo.addrs())) == "(Hex('0x28'), Hex('0x30'))"

    wordinfo = ua.WordInfo.create(addrmap, item, 5, mask=0xF, offset=8)
    assert str(wordinfo) == "0x20 [0x28 1x32 (4 bytes)] mask=0xF SINGLE: 5"
    assert str(tuple(wordinfo.iter())) == "((Hex('0x28'), 5),)"


def test_addrspace(addrmap):
    """Address Space."""
    wordinfo = ua.WordInfo.create(addrmap, "spi", 5)
    assert str(wordinfo) == "spi [0x20 1x32 (4 bytes)] SINGLE: 5"
    assert str(tuple(wordinfo.iter())) == "((Hex('0x20'), 5),)"

    wordinfo = ua.WordInfo.create(addrmap, "spi", (5, 6))
    assert str(wordinfo) == "spi [0x20 2x32 (8 bytes)] BURST: (5, 6)"
    assert str(tuple(wordinfo.iter())) == "((Hex('0x20'), 5), (Hex('0x24'), 6))"

    wordinfo = ua.WordInfo.create(addrmap, "spi", ((8, 5), (16, 6)))
    assert str(wordinfo) == "spi [0x20 5x32 (20 bytes)] SCAT: ((8, 5), (16, 6))"
    assert str(tuple(wordinfo.iter())) == "((Hex('0x28'), 5), (Hex('0x30'), 6))"

    wordinfo = ua.WordInfo.create(addrmap, "spi", 5, mask=0xF, offset=8)
    assert str(wordinfo) == "spi [0x28 1x32 (4 bytes)] mask=0xF SINGLE: 5"
    assert str(tuple(wordinfo.iter())) == "((Hex('0x28'), 5),)"

    with raises(ValueError, match=re.escape("offset=32 exceeds size (32 bytes)")):
        ua.WordInfo.create(addrmap, "spi", 5, offset=32)

    with raises(ValueError, match=re.escape("data size 128 exceeds size (32 bytes)")):
        ua.WordInfo.create(addrmap, "spi", tuple(range(32)))

    with raises(ValueError, match=re.escape("'wordsize' is forbidden for non integer")):
        ua.WordInfo.create(addrmap, "spi", 5, wordsize=8)


def test_word(addrmap):
    """Word."""
    wordinfo = ua.WordInfo.create(addrmap, "spi.ctrl", "ena=ena")
    assert str(wordinfo) == "spi.ctrl [0x20 1x32 (4 bytes)] mask=0x1 RMW SINGLE: 1"

    wordinfo = ua.WordInfo.create(addrmap, "spi.ctrl", "ena=ena, mode=9")
    assert str(wordinfo) == "spi.ctrl [0x20 1x32 (4 bytes)] mask=0x1F RMW SINGLE: 19"

    wordinfo = ua.WordInfo.create(addrmap, "spi.ctrl", "ena=ena, mode=9, speed=0x14")
    assert str(wordinfo) == "spi.ctrl [0x20 1x32 (4 bytes)] mask=0x1FFF SINGLE: 659"

    wordinfo = ua.WordInfo.create(addrmap, "spi.ctrl", "ena=ena, *=2")
    assert str(wordinfo) == "spi.ctrl [0x20 1x32 (4 bytes)] mask=0x1FFF SINGLE: 69"

    wordinfo = ua.WordInfo.create(addrmap, "spi.ctrl", "*=<RST>")
    assert str(wordinfo) == "spi.ctrl [0x20 1x32 (4 bytes)] mask=0x1FFF SINGLE: 98"

    wordinfo = ua.WordInfo.create(addrmap, "spi.rc", "bit=1")
    assert str(wordinfo) == "spi.rc [0x28 1x32 (4 bytes)] mask=0x1 !R SINGLE: 1"

    wordinfo = ua.WordInfo.create(addrmap, "spi.wl", "bit=1")
    assert str(wordinfo) == "spi.wl [0x2C 1x32 (4 bytes)] mask=0x1 !W SINGLE: 1"

    with raises(ValueError, match=re.escape("Value 'disenabled' is not covered by EnaType()")):
        ua.WordInfo.create(addrmap, "spi.ctrl", "ena=disenabled")

    wordinfo = ua.WordInfo.create(addrmap, "spi.ctrl", {"ena": 1})
    assert str(wordinfo) == "spi.ctrl [0x20 1x32 (4 bytes)] mask=0x1 RMW SINGLE: 1"

    with raises(TypeError, match=re.escape("'data' must be 'int', 'str' or 'dict' for fields, not (1, 2)")):
        ua.WordInfo.create(addrmap, "spi.ctrl", (1, 2))

    wordinfo = ua.WordInfo.create(addrmap, "spi.ctrl", 8)
    assert str(wordinfo) == "spi.ctrl [0x20 1x32 (4 bytes)] SINGLE: 8"

    with raises(ValueError, match=re.escape("Unknown field names: 'foo', 'bar'")):
        ua.WordInfo.create(addrmap, "spi.ctrl", {"ena": 1, "foo": 3, "bar": 5})


def test_field(addrmap):
    """Field."""
    wordinfo = ua.WordInfo.create(addrmap, "spi.ctrl.ena", "ena")
    assert str(wordinfo) == "spi.ctrl.ena [0x20 1x32 (4 bytes)] mask=0x1 RMW SINGLE: 1"

    wordinfo = ua.WordInfo.create(addrmap, "spi.ctrl.ena", "dis")
    assert str(wordinfo) == "spi.ctrl.ena [0x20 1x32 (4 bytes)] mask=0x1 RMW SINGLE: 0"

    wordinfo = ua.WordInfo.create(addrmap, "spi.ctrl.ena", 1)
    assert str(wordinfo) == "spi.ctrl.ena [0x20 1x32 (4 bytes)] mask=0x1 RMW SINGLE: 1"

    wordinfo = ua.WordInfo.create(addrmap, "spi.ctrl.mode", 9)
    assert str(wordinfo) == "spi.ctrl.mode [0x20 1x32 (4 bytes)] mask=0x1E RMW SINGLE: 18"

    with raises(ValueError, match=re.escape("'offset' is not allowed for words and fields")):
        ua.WordInfo.create(addrmap, "spi.ctrl.mode", 5, offset=32)

    with raises(ValueError, match=re.escape("'mask' is not allowed for fields")):
        ua.WordInfo.create(addrmap, "spi.ctrl.mode", 5, mask=31)

    with raises(TypeError, match=re.escape("'data' must be 'int' or 'str' for fields, not (5, 6)")):
        ua.WordInfo.create(addrmap, "spi.ctrl.mode", (5, 6))

    with raises(TypeError, match=re.escape("'data' must be 'int' or 'str' for fields, not ((8, 5), (16, 6))")):
        ua.WordInfo.create(addrmap, "spi.ctrl.mode", ((8, 5), (16, 6)))



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Utilities.
"""

import ucdp as u
from humannum import bytesize_


def calc_depth_size(width: int, depth: int | None = None, size: u.Bytes | None = None) -> tuple[int, u.Bytes]:
    """
    Calc Either Depth Or Size.

    >>> calc_depth_size(32, depth=64)
    (64, Bytesize('256 bytes'))
    >>> calc_depth_size(32, size='64')
    (16, Bytesize('64 bytes'))
    >>> calc_depth_size(32, depth=64, size=256)
    (64, Bytesize('256 bytes'))
    >>> calc_depth_size(32)
    Traceback (most recent call last):
        ...
    ValueError: Either 'depth' or 'size' are required.
    >>> calc_depth_size(32, depth=64, size='64')
    Traceback (most recent call last):
        ...
    ValueError: 'depth' and 'size' are mutually exclusive.
    """
    # Provide either 'size' or 'depth' and calculate the other
    if size is None:
        if depth is None:
            raise ValueError("Either 'depth' or 'size' are required.")
        size = bytesize_((width * depth) // 8)
    else:
        size = bytesize_(size)
        depth_calculated = int(size * 8 // width)
        if depth is not None and depth != depth_calculated:
            raise ValueError("'depth' and 'size' are mutually exclusive.")
        depth = depth_calculated
    return depth, size


def split_keyvaluepairs(value: str) -> dict[str, str]:
    """
    Split comma-separated fields into (name, value) pairs.

    >>> split_keyvaluepairs('a=4')
    {'a': '4'}
    >>> split_keyvaluepairs('a=4, b=a')
    {'a': '4', 'b': 'a'}
    >>> split_keyvaluepairs('a=4, b')
    Traceback (most recent call last):
      ...
    ValueError: Invalid key=value pair: 'b'
    >>> split_keyvaluepairs('a=4, a=a')
    Traceback (most recent call last):
      ...
    ValueError: Duplicate key 'a'
    """
    data = {}
    for pair in value.split(","):
        pair = pair.strip()  # noqa: PLW2901
        if "=" not in pair:
            raise ValueError(f"Invalid key=value pair: '{pair}'")
        fname, fvalue = pair.split("=", 1)
        if fname not in data:
            data[fname] = fvalue
        else:
            raise ValueError(f"Duplicate key {fname!r}")
    return data



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Address Map.
"""

from collections.abc import Callable, Iterator
from typing import TypeAlias

import aligntext
import ucdp as u
from icdutil import num
from ucdp_glbl.attrs import format_attrs

from .addrmapref import AddrMapRef
from .addrspace import Addrspace, ReservedAddrspace
from .addrspaces import Addrspaces
from .defines import Defines

FillAddrspaceFactory: TypeAlias = Callable[[int, int, int], Addrspace]
AddrspaceFilter: TypeAlias = Callable[[Addrspace], bool]


class AddrMap(u.Object):
    """Address Map."""

    unique: bool = False
    fixed_size: u.Bytes | None = None
    defines: Defines = u.Field(default_factory=dict)
    ref: u.TopModRef | None = None

    @staticmethod
    def from_addrspaces(
        addrspaces: Addrspaces,
        unique: bool = False,
        fixed_size: u.Bytes | None = None,
        defines: Defines | None = None,
        ref: u.TopModRef | None = None,
    ) -> "AddrMap":
        """Create From address spaces."""
        addrmap = AddrMap(unique=unique, fixed_size=fixed_size, defines=defines or {}, ref=ref)
        for addrspace in addrspaces:
            addrmap.add(addrspace)
        return addrmap

    _addrspaces: list[Addrspace] = u.PrivateField(default_factory=list)

    def add(self, addrspace: Addrspace) -> None:
        """Add Address Space."""
        self._check_size(addrspace)

        pos = self._find_pos(addrspace)

        if self.unique:
            self._check_overlapping(pos, addrspace)

        self._addrspaces.insert(pos, addrspace)

    def get(self, addrspace: Addrspace) -> AddrMapRef:
        """Add Address Space."""
        for item in self._addrspaces:
            intersect = item.get_intersect(addrspace)
            if intersect:
                return AddrMapRef.create(addrspace=intersect)
        raise ValueError(f"{addrspace!r} does not exists")

    def __iter__(self) -> Iterator[Addrspace]:
        yield from self._addrspaces

    def iter(  # noqa: C901
        self,
        filter_: AddrspaceFilter | None = None,
        fill: FillAddrspaceFactory | bool | None = None,
    ) -> Iterator[Addrspace]:
        """Iterate over Address Spaces."""

        def no_addrspacefilter(_: Addrspace) -> bool:
            return True

        if fill is True:
            fill = create_fill_addrspace

        if not fill:
            if filter_ is None:
                yield from self._addrspaces
            else:
                for addrspace in self._addrspaces:
                    if filter_(addrspace):
                        yield addrspace
        else:
            filter_ = filter_ or no_addrspacefilter
            baseaddr = 0
            idx = 0
            for addrspace in self._addrspaces:
                if not filter_(addrspace):
                    continue

                # fill before
                if addrspace.baseaddr > baseaddr:
                    yield fill(idx, baseaddr, addrspace.baseaddr - baseaddr)
                    idx += 1

                # element
                yield addrspace
                baseaddr = addrspace.nextaddr

            if self.addrwidth:
                nextaddr = 2**self.addrwidth

                # fill end
                if nextaddr > baseaddr:
                    yield fill(idx, baseaddr, nextaddr - baseaddr)

    @property
    def size(self) -> u.Bytes | None:
        """
        Size in Bytes.
        """
        size = self.fixed_size
        if size is None:
            lastaddr = self.lastaddr
            if lastaddr is None:
                return None
            size = u.Bytesize(lastaddr + 1)
        return size

    @property
    def addrwidth(self) -> int | None:
        """
        Address Width.
        """
        size = self.size
        if size is None:
            return None
        return num.calc_unsigned_width(int(size - 1))

    @property
    def firstaddr(self) -> int | None:
        """
        First used address.
        """
        try:
            return self._addrspaces[0].baseaddr
        except IndexError:
            return None

    @property
    def lastaddr(self) -> int | None:
        """
        Last used address.
        """
        try:
            return self._addrspaces[-1].endaddr
        except IndexError:
            return None

    @property
    def addrslice(self) -> u.Slice | None:
        """
        Address Slice.
        """
        addrspaces = self._addrspaces
        addrwidth = self.addrwidth
        if addrwidth is None or not addrspaces:
            return None

        minsize = int(min(addrspace.size for addrspace in addrspaces))

        # Ensure at least one address decoding bit
        if minsize == self.size:
            left = addrwidth
        else:
            left = addrwidth - 1

        right = min(num.calc_lowest_bit_set(minsize), left)
        return u.Slice(left=left, right=right)

    def get_free_baseaddr(self, size: u.Bytes, align=None, start=None) -> int:
        """
        Return baseaddress of free window with `size`.

        Args:
            size: Window Size

        Keyword Args:
            align: Alignment, default aligned to size
            start: Start search behind given address
        """
        size = u.Bytesize(size)
        if align is None:
            align = size
        if start is None:
            lastaddr = self.lastaddr
            start = lastaddr + 1 if lastaddr else 0

        baseaddr = u.Hex(self._find_space(size, align, start))

        # End Check
        endaddr = baseaddr + size
        size = self.fixed_size
        if size is not None:
            if endaddr >= size:
                size_hex = u.Hex(size)
                raise ValueError(f"End address {endaddr} would exceed maximum size {size_hex} ({size})")

        return baseaddr

    def _find_space(self, size, align, start):
        addr = start = num.align(start, align=align)
        for item in self._addrspaces:
            if item.endaddr < start:
                # skip all before start
                addr = max(num.align(item.nextaddr, align=align), addr)
                continue
            if item.baseaddr >= (addr + size):
                break
            addr = num.align(item.nextaddr, align=align)
        return addr

    def get_overview(self, minimal: bool = False, fill: FillAddrspaceFactory | bool | None = None) -> str:
        """
        Return overview table.
        """

        def align(data) -> str:
            lines = list(data)
            lens = (max(len(cell) for cell in row) for row in zip(*lines, strict=False))
            lines.insert(1, ("-" * len_ for len_ in lens))
            return aligntext.align(lines, seps=(" | ",), sepfirst="| ", seplast=" |") + "\n"

        defines = ", ".join(f"{define}={value!r}" for define, value in self.defines.items()) or None
        header = [
            f"* Top:     `{self.ref}`",
            f"* Defines: `{defines}`",
            f"* Size:    `{self.size}`",
        ]
        parts = [
            "\n".join(header),
            align(self.get_addrspaces_overview(fill=fill)),
        ]
        if not minimal:
            parts.append(align(self.get_word_fields_overview()))
        return "\n\n".join(parts)

    def get_addrspaces_overview(
        self, fill: FillAddrspaceFactory | bool | None = None
    ) -> Iterator[tuple[str, str, str, str, str]]:
        """Get Address Spaces Overview Data."""
        yield ("Addrspace", "Type", "Base", "Size", "Infos", "Attributes")
        for addrspace in self.iter(fill=fill):
            classname = addrspace.__class__.__name__.replace("Addrspace", "") or "-"
            infos = []
            if addrspace.is_sub:
                infos.append("Sub")
            if addrspace.is_volatile:
                infos.append("Volatile")
            yield (
                addrspace.name,
                classname,
                f"`{addrspace.base}`",
                f"`{addrspace.org}`",
                ",".join(infos),
                format_attrs(addrspace.attrs),
            )

    def get_word_fields_overview(  # noqa: C901
        self, addrspaces: bool = True, words: bool = True, fields: bool = True
    ) -> Iterator[tuple[str, str, str, str, str, str, str]]:
        """Get Word-Fields Overview Data."""
        yield ("Addrspace", "Word", "Field", "Offset", "Access", "Reset", "Infos", "Attributes")
        resolver = u.ExprResolver()
        for addrspace in self:
            # address spaces
            if addrspaces:
                infos = []
                if addrspace.is_volatile:
                    infos.append("Volatile")
                yield (
                    addrspace.name,
                    "",
                    "",
                    f"`{addrspace.base}`",
                    addrspace.access,
                    "",
                    ",".join(infos),
                    format_attrs(addrspace.attrs),
                )
            # words
            for word in addrspace.words:
                if words:
                    infos = []
                    if word.is_volatile:
                        infos.append("Volatile")
                    yield (
                        addrspace.name,
                        word.name,
                        "",
                        f"`  +{word.slice}`",
                        word.access,
                        "",
                        ",".join(infos),
                        format_attrs(word.attrs),
                    )
                # fields
                if fields:
                    for field in word.fields:
                        infos = []
                        if field.is_volatile:
                            infos.append("Volatile")
                        if field.is_const:
                            infos.append("CONST")
                        reset = resolver.resolve_value(field.type_)
                        yield (
                            addrspace.name,
                            word.name,
                            field.name,
                            f"`    [{field.slice}]`",
                            str(field.access),
                            f"`{reset}`",
                            ",".join(infos),
                            format_attrs(field.attrs),
                        )

    def _check_size(self, addrspace: Addrspace) -> None:
        fixed_size = self.fixed_size
        if fixed_size is not None:
            if addrspace.endaddr >= fixed_size:
                raise ValueError(f"{addrspace!r}: exceeds maximum size: {fixed_size}.")

    def _find_pos(self, addrspace: Addrspace) -> int:
        baseaddr = addrspace.baseaddr
        for pos, item in enumerate(self._addrspaces):
            if item.baseaddr > baseaddr:
                return pos
        return len(self._addrspaces)

    def _check_overlapping(self, pos: int, addrspace: Addrspace) -> None:
        addrspaces = self._addrspaces
        # lower
        try:
            lower = addrspaces[pos - 1]
        except IndexError:
            pass
        else:
            if addrspace.is_overlapping(lower):
                raise ValueError(f"{addrspace!r} overlaps with {lower!r}")
        # upper
        try:
            upper = addrspaces[pos]
        except IndexError:
            pass
        else:
            if addrspace.is_overlapping(upper):
                raise ValueError(f"{addrspace!r} overlaps with {upper!r}")


def create_fill_addrspace(idx, baseaddr, size) -> Addrspace:
    """Create Fill Addrspace."""
    return ReservedAddrspace(name=f"reserved{idx}", baseaddr=baseaddr, size=size, is_sub=False)



"""Pytest Configuration and Fixtures."""

from pathlib import Path

import ucdp as u
from pytest import fixture

import ucdp_addr


@fixture
def template_path():
    """Path to templates."""
    return Path(ucdp_addr.__file__).parent / "templates"


@fixture
def testdata_path():
    """Register Testdata."""
    path = Path(__file__).parent / "testdata"
    with u.extend_sys_path([path]):
        yield path



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Address Space."""

import re

import ucdp as u
from pytest import fixture, raises
from test2ref import assert_refdata

from ucdp_addr import AddrMap, Addrspace, AddrspaceAlias, create_fill_addrspace


@fixture
def zero():
    """Zero."""
    return Addrspace(name="zero", baseaddr=0x0000, size=0x1000)


@fixture
def one():
    """One."""
    addrspace = Addrspace(name="one", baseaddr=0x1000, size=0x1000, is_volatile=True, attrs="a=1;b")
    word = addrspace.add_word("word0")
    word.add_field("field", u.UintType(3), "RW", attrs="foo")
    word = addrspace.add_word("word1", depth=4, attrs="bar=4")
    word.add_field("field", u.UintType(3), "RW")
    return addrspace


@fixture
def two():
    """Two."""
    addrspace = Addrspace(name="two", baseaddr=0x3000, size=0x1000)
    word = addrspace.add_word("word", offset=8)
    word.add_field("field", u.UintType(3), "RW", offset=6)
    return addrspace


@fixture
def three():
    """Three."""
    return Addrspace(name="three", baseaddr=0x7F000000, size=0x1000, is_sub=False)


@fixture
def four():
    """Four."""
    # Overlapping with two at the end
    return Addrspace(name="four", baseaddr=0x3800, size=0x800)


@fixture
def five():
    """Five."""
    # Overlapping with two at the start
    return Addrspace(name="five", baseaddr=0x2E00, size=0x400)


@fixture
def alias(one):
    """Alias to one."""
    return AddrspaceAlias(one, baseaddr=0x6000)


def test_empty(tmp_path):
    """Empty Address Map."""
    addrmap = AddrMap()
    assert repr(addrmap) == "AddrMap()"

    assert addrmap.size is None
    assert addrmap.addrwidth is None
    assert addrmap.firstaddr is None
    assert addrmap.lastaddr is None
    assert addrmap.addrslice is None

    (tmp_path / "overview.md").write_text(addrmap.get_overview())
    assert_refdata(test_empty, tmp_path)


def test_growing(tmp_path, one, two, three, four, alias):
    """Growing Address Map."""
    addrmap = AddrMap()
    assert repr(addrmap) == "AddrMap()"

    addrmap.add(one)
    assert repr(addrmap.size) == "Bytesize('8 KB')"
    assert addrmap.addrwidth == 13
    assert addrmap.addrslice.bits == "12"
    assert repr(addrmap.firstaddr) == "Hex('0x1000')"
    assert repr(addrmap.lastaddr) == "Hex('0x1FFF')"

    addrmap.add(two)
    assert repr(addrmap.size) == "Bytesize('16 KB')"
    assert addrmap.addrwidth == 14
    assert addrmap.addrslice.bits == "13:12"
    assert repr(addrmap.firstaddr) == "Hex('0x1000')"
    assert repr(addrmap.lastaddr) == "Hex('0x3FFF')"

    addrmap.add(three)
    assert repr(addrmap.size) == "Bytesize('2080772 KB')"
    assert addrmap.addrwidth == 31
    assert addrmap.addrslice.bits == "30:12"
    assert repr(addrmap.firstaddr) == "Hex('0x1000')"
    assert repr(addrmap.lastaddr) == "Hex('0x7F000FFF')"

    addrmap.add(alias)
    addrmap.add(four)

    assert tuple(addrmap) == (one, two, four, alias, three)
    assert tuple(addrmap.iter()) == (one, two, four, alias, three)
    assert tuple(addrmap.iter(filter_=lambda addrspace: addrspace.name != "two")) == (one, four, alias, three)

    with (tmp_path / "iter-fillfunc.txt").open("w") as file:
        for item in addrmap.iter(fill=create_fill_addrspace):
            file.write(f"{item!r}\n")

    with (tmp_path / "iter-filltrue.txt").open("w") as file:
        for item in addrmap.iter(fill=True):
            file.write(f"{item!r}\n")

    with (tmp_path / "iter-fill-filter.txt").open("w") as file:
        for item in addrmap.iter(fill=True, filter_=lambda addrspace: addrspace.name != "two"):
            file.write(f"{item!r}\n")

    (tmp_path / "overview.md").write_text(addrmap.get_overview())
    assert_refdata(test_growing, tmp_path)


def test_fixed_size(tmp_path, one, two, three, alias):
    """Growing Address Map."""
    addrmap = AddrMap(fixed_size=0xF0000)
    assert repr(addrmap) == "AddrMap(fixed_size=Bytesize('960 KB'))"

    assert repr(addrmap.size) == "Bytesize('960 KB')"
    assert addrmap.addrwidth == 20
    assert addrmap.addrslice is None
    assert addrmap.firstaddr is None
    assert addrmap.lastaddr is None

    addrmap.add(one)
    assert repr(addrmap.size) == "Bytesize('960 KB')"
    assert addrmap.addrwidth == 20
    assert addrmap.addrslice.bits == "19:12"
    assert repr(addrmap.firstaddr) == "Hex('0x1000')"
    assert repr(addrmap.lastaddr) == "Hex('0x1FFF')"

    addrmap.add(two)
    assert repr(addrmap.size) == "Bytesize('960 KB')"
    assert addrmap.addrwidth == 20
    assert addrmap.addrslice.bits == "19:12"
    assert repr(addrmap.firstaddr) == "Hex('0x1000')"
    assert repr(addrmap.lastaddr) == "Hex('0x3FFF')"

    msg = "size=Bytesize('4 KB'), is_sub=False): exceeds maximum size: 960 KB."
    with raises(ValueError, match=re.escape(msg)):
        addrmap.add(three)
    assert repr(addrmap.size) == "Bytesize('960 KB')"
    assert addrmap.addrwidth == 20
    assert addrmap.addrslice.bits == "19:12"
    assert repr(addrmap.firstaddr) == "Hex('0x1000')"
    assert repr(addrmap.lastaddr) == "Hex('0x3FFF')"

    addrmap.add(alias)

    assert tuple(addrmap) == (one, two, alias)

    (tmp_path / "overview.md").write_text(addrmap.get_overview())
    assert_refdata(test_fixed_size, tmp_path)


def test_unique(tmp_path, one, two, three, four, five):
    """Unique Address Map."""
    addrmap = AddrMap(unique=True)
    addrmap.add(one)
    addrmap.add(two)
    addrmap.add(three)

    msg = "size=Bytesize('2 KB')) overlaps with Addrspace(name='two',"
    with raises(ValueError, match=re.escape(msg)):
        addrmap.add(four)

    msg = "size=Bytesize('1 KB')) overlaps with Addrspace(name='two', baseaddr=Hex('0x3000')"
    with raises(ValueError, match=re.escape(msg)):
        addrmap.add(five)

    (tmp_path / "overview.md").write_text(addrmap.get_overview())
    assert_refdata(test_unique, tmp_path)


def test_corner(tmp_path, zero, one):
    """Corner Case."""
    addrmap = AddrMap()
    addrmap.add(zero)
    assert addrmap.addrslice.bits == "12"
    addrmap.add(one)
    assert addrmap.addrslice.bits == "12"

    (tmp_path / "overview.md").write_text(addrmap.get_overview())
    assert_refdata(test_corner, tmp_path)


def test_get_free_baseaddr(tmp_path, one, two, three):
    """Test Get Free Base Address."""
    addrmap = AddrMap(fixed_size=0x8000)

    # empty
    assert addrmap.get_free_baseaddr(1) == 0
    assert addrmap.get_free_baseaddr(0x1000) == 0
    assert addrmap.get_free_baseaddr(0x1000, align=0x2000) == 0
    assert addrmap.get_free_baseaddr(0x1000, align=0x2000, start=0x800) == 0x2000
    assert addrmap.get_free_baseaddr(0x1000, align=0x2000, start=0x3000) == 0x4000
    assert addrmap.get_free_baseaddr(0x1000, align=0x2000, start=0x4000) == 0x4000
    assert addrmap.get_free_baseaddr(0x1000, start=0x800) == 0x1000
    assert addrmap.get_free_baseaddr(0x1000, start=0x3000) == 0x3000

    # one element
    addrmap.add(one)

    assert addrmap.get_free_baseaddr(1) == 0x2000
    assert addrmap.get_free_baseaddr(0x1000) == 0x2000
    assert addrmap.get_free_baseaddr(0x1000, align=0x800) == 0x2000
    assert addrmap.get_free_baseaddr(0x1000, align=0x800, start=0x800) == 0x2000
    assert addrmap.get_free_baseaddr(0x1000, align=0x800, start=0x3000) == 0x3000
    assert addrmap.get_free_baseaddr(0x1000, align=0x800, start=0x4000) == 0x4000
    assert addrmap.get_free_baseaddr(0x1000, start=0x800) == 0x2000
    assert addrmap.get_free_baseaddr(0x1000, start=0x3000) == 0x3000

    # second element
    addrmap.add(two)

    assert addrmap.get_free_baseaddr(1) == 0x4000
    assert addrmap.get_free_baseaddr(0x1000) == 0x4000
    assert addrmap.get_free_baseaddr(0x1000, align=0x800) == 0x4000
    assert addrmap.get_free_baseaddr(0x1000, align=0x800, start=0x800) == 0x2000
    assert addrmap.get_free_baseaddr(0x1000, align=0x800, start=0x3000) == 0x4000
    assert addrmap.get_free_baseaddr(0x1000, align=0x800, start=0x4000) == 0x4000
    assert addrmap.get_free_baseaddr(0x1000, start=0x800) == 0x2000

    msg = "End address 0x9000 would exceed maximum size 0x8000 (32 KB)"
    with raises(ValueError, match=re.escape(msg)):
        addrmap.get_free_baseaddr(0x1000, start=0x8000)



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Master-Slave Matrix.
"""

from collections import defaultdict
from logging import getLogger

import aligntext
import ucdp as u

from .addrdecoder import AddrDecoder
from .addrmaster import AddrMaster
from .addrslave import AddrSlave

LOGGER = getLogger(__name__)


class AddrMatrix(AddrDecoder):
    """
    Bus Matrix with a number of masters and slave with optional connections.
    """

    masters: u.Namespace = u.Field(default_factory=u.Namespace)
    """Masters."""

    _master_slaves = u.PrivateField(default_factory=lambda: defaultdict(list))
    _slave_masters = u.PrivateField(default_factory=lambda: defaultdict(list))

    def _add_master(
        self,
        master: AddrMaster,
        slavenames: u.Names | None = None,
    ):
        self._check_lock()
        self.masters.add(master)
        self.add_interconnects((master.name,), slavenames)

    def _add_slave(
        self,
        slave: AddrSlave,
        masternames: u.Names | None = None,
        baseaddr=u.AUTO,
        size: u.Bytes | None = None,
    ):
        self._check_lock()
        self.slaves.add(slave)
        if baseaddr is not None and (size is not None or self.default_size):
            slave.add_addrrange(baseaddr, size)
        self.add_interconnects(masternames, (slave.name,))

    def add_interconnects(self, masternames: u.Names, slavenames: u.Names):
        """Add Interconnects."""
        self._check_lock()
        for mastername in u.split(masternames):
            for slavename in u.split(slavenames):
                if slavename in self._master_slaves[mastername]:
                    LOGGER.warning("%s: interconnect %s -> %s already exists", self, mastername, slavename)
                    continue
                self._master_slaves[mastername].append(slavename)
                self._slave_masters[slavename].append(mastername)

    @property
    def master_slaves(self) -> tuple[tuple[AddrMaster, tuple[AddrSlave, ...]], ...]:
        """Masters and Their Slaves."""
        pairs: list[tuple[AddrMaster, tuple[AddrSlave, ...]]] = []
        slaves = self.slaves
        for master in self.masters:
            slavenames = self._master_slaves[master.name]
            masterslaves = tuple(slaves[name] for name in slavenames if name in slaves)
            pairs.append((master, masterslaves))
        return tuple(pairs)

    @property
    def slave_masters(self) -> tuple[tuple[AddrSlave, tuple[AddrMaster, ...]], ...]:
        """Slaves and Their Masters."""
        pairs: list[tuple[AddrSlave, tuple[AddrMaster, ...]]] = []
        masters = self.masters
        for slave in self.slaves:
            masternames = self._slave_masters[slave.name]
            masters = tuple(masters[name] for name in masternames if name in masters)
            pairs.append((slave, masters))
        return tuple(pairs)

    def _check_lock(self):
        """Check Lock."""

    def _check_masters_slaves(self):
        """Run Consistency Checks."""
        masters = self.masters
        slaves = self.slaves
        if not masters:
            LOGGER.warning("%s has no masters", self)
        if not slaves:
            LOGGER.warning("%s has no slaves", self)
        for master, slaves in self.master_slaves:
            if not slaves:
                LOGGER.warning("%s: %r has no slaves", self, master)
        for slave, masters in self.slave_masters:
            if not masters:
                LOGGER.warning("%s: %r has no masters", self, slave)
        for unknown_master in set(self._master_slaves.keys()) - set(self.masters.keys()):
            LOGGER.warning("Master %r is not known", unknown_master)
        for unknown_slave in set(self._slave_masters.keys()) - set(self.slaves.keys()):
            LOGGER.warning("Slaves %r is not known", unknown_slave)

    def get_overview(self):
        """Return overview tables."""
        overview = [
            self._get_overview_matrix(),
            self.addrmap.get_overview(minimal=True, fill=True),
        ]
        return "\n\n\n".join(overview)

    def _get_overview_matrix(self) -> str:
        def align(data) -> str:
            lines = list(data)
            lens = (max(len(cell) for cell in row) for row in zip(*lines, strict=False))
            lines.insert(1, ("-" * len_ for len_ in lens))
            return aligntext.align(lines, seps=(" | ",), sepfirst="| ", seplast=" |") + "\n"

        slaves = self.slaves
        headers = ["Master > Slave"] + [slave.name for slave in slaves]
        matrix = [headers]
        idxmap = {slave.name: idx for idx, slave in enumerate(slaves, 1)}
        empty = ["" for slave in slaves]
        for master, slaves in self.master_slaves:
            item = [master.name, *empty]
            for slave in slaves:
                item[idxmap[slave.name]] = "X"
            matrix.append(item)

        return align(matrix)



"""
Word Information.

"""

from collections.abc import Iterator
from typing import TypeAlias

import ucdp as u
from pydantic import PositiveInt

from .access import is_read_repeatable, is_write_repeatable
from .addrinfo import _apply_offset
from .addrmap import AddrMap
from .addrmapref import AddrMapRef, ToAddrMapRef
from .addrrange import AddrRange
from .addrspace import Word, read_on_modify, resolve_field_value
from .data import Data, DataType, get_datatype, get_size
from .datainfo import AddrData, _iter_data
from .resolver import resolve
from .util import split_keyvaluepairs

Offset: TypeAlias = PositiveInt
Size: TypeAlias = PositiveInt
Depth: TypeAlias = PositiveInt
FieldValues: TypeAlias = dict[str, str | int]


class WordInfo(u.Object):
    """WordInfo Info."""

    ref: AddrMapRef
    """Address Map Reference."""
    addrrange: AddrRange
    """Accessed Address Range."""
    datatype: DataType
    """Data Type."""
    data: Data
    """Data."""
    mask: u.Hex | None = None
    """Data Masking."""
    rmw: bool | None = None
    """Read Modify Write - Read before write required."""
    read_repeatable: bool | None = None
    """Read Operation Has No Side-Effects."""
    write_repeatable: bool | None = None
    """Write Operation Has No Side-Effects."""

    def __str__(self):
        info = f"{self.ref!s} [{self.addrrange!s}]"
        if self.mask is not None:
            info = f"{info} mask={self.mask}"
        if self.rmw:
            info = f"{info} RMW"
        if self.read_repeatable is False:
            info = f"{info} !R"
        if self.write_repeatable is False:
            info = f"{info} !W"
        return f"{info} {self.datatype.name}: {self.data}"

    @staticmethod
    def create(
        addrmap: AddrMap,
        item: ToAddrMapRef,
        data: Data | str | FieldValues,
        offset: Offset | None = None,
        mask: int | None = None,
        wordsize: Size | None = None,
    ) -> "WordInfo":
        """
        Create `WordInfo`.

        Args:
            addrmap: Address Map
            item: Thing to be resolved
            data: Data

        Keyword Args:
            offset: address offset in bytes
            mask: Mask for non-field access.
            wordsize: Size of one word in bytes (Only allowed if `item` is an `int`)
        """
        # determine size from data on raw integer address
        if isinstance(item, int):
            item = AddrRange(baseaddr=item + (offset or 0), size=get_size(data, wordsize=wordsize or 4))
            offset = None
        elif wordsize is not None:
            raise ValueError("'wordsize' is forbidden for non integer.")

        ref = resolve(addrmap, item)
        addrrange = ref.addrrange

        # apply offset
        if offset is not None:
            addrrange = _apply_offset(ref, addrrange, offset)

        if ref.word:
            if ref.field:
                # field
                if mask is not None:
                    raise ValueError("'mask' is not allowed for fields")
                if not isinstance(data, (int, str)):
                    raise TypeError(f"'data' must be 'int' or 'str' for fields, not {data}")
                slice_ = ref.field.slice
                data = slice_.update(0, resolve_field_value(ref.field, data))
                mask = slice_.mask
            else:
                # word
                if not isinstance(data, (int, str, dict)):
                    raise TypeError(f"'data' must be 'int', 'str' or 'dict' for fields, not {data}")
                data, mask = _resolve_word(ref.word, data, mask)
        else:
            # address
            datatype = get_datatype(data)
            addrrange = _strip_size(data, datatype, addrrange)  # type: ignore[arg-type]
            return WordInfo(
                ref=ref,
                datatype=datatype,
                addrrange=addrrange,
                data=data,
                mask=mask,
            )
        accesses = [field.bus for field in ref.word.fields if field.bus]
        return WordInfo(
            ref=ref,
            addrrange=addrrange,
            data=data,
            datatype=DataType.SINGLE,
            mask=mask,
            rmw=read_on_modify(ref.word, mask) if mask is not None else False,
            read_repeatable=is_read_repeatable(accesses),
            write_repeatable=is_write_repeatable(accesses),
        )

    def iter(self) -> Iterator[AddrData]:
        """Iteratate over single address value pairs according to access."""
        yield from _iter_data(self.datatype, self.data, self.addrrange)

    def addrs(self) -> Iterator[int]:
        """Addresses."""
        for addr, _ in _iter_data(self.datatype, self.data, self.addrrange):
            yield addr


def _strip_size(data: Data, datatype: DataType, addrrange: AddrRange) -> AddrRange:
    if datatype == DataType.SINGLE:
        size = addrrange.wordsize
    elif datatype == DataType.BURST:
        size = int(addrrange.wordsize * len(data))  # type: ignore[operator,arg-type]
    else:
        addrs = [addr for addr, _ in data]  # type:ignore[union-attr]
        size = max(addrs) + addrrange.wordsize

    if size > addrrange.size:
        raise ValueError(f"data size {size!r} exceeds size ({addrrange.size})")
    return addrrange.new(size=size, depth=None)


def _resolve_word(word: Word, data: str | int | FieldValues, mask: int | None) -> tuple[int, int | None]:
    if isinstance(data, int):
        return data, mask

    # str ==> dict
    if isinstance(data, str):
        values = split_keyvaluepairs(data)
    else:
        values = dict(data)  # type: ignore[arg-type]

    # Default
    default = values.pop("*", None)
    if default:
        for field in word.fields:
            if field.name not in values:
                values[field.name] = resolve_field_value(field, default)  # type: ignore[assignment]

    # Evaluate Fields
    result, mask = 0, 0
    for field in word.fields:
        try:
            rawvalue = values.pop(field.name)
        except KeyError:
            continue
        value = resolve_field_value(field, rawvalue)
        slice_ = field.slice
        result = slice_.update(result, value, is_signed=u.is_signed(field.type_))
        mask |= slice_.mask

    # Unknown ones
    if values:
        fieldnames = ", ".join(repr(value) for value in values)
        raise ValueError(f"Unknown field names: {fieldnames}")
    return result, mask



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Address Decoder."""

import re

import ucdp as u
from pytest import raises
from test2ref import assert_refdata

from ucdp_addr import NOREF, AddrDecoder, AddrMap, AddrSlave, Addrspace, Addrspaces


class Mod(u.AMod):
    """Just a module."""

    def _build(self) -> None:
        pass

    def get_addrspaces(self, master=None, **kwargs) -> Addrspaces:
        """Address Spaces."""
        addrspace = Addrspace(name=self.hiername, width=32, depth=64)
        addrspace.add_word("one")
        if not master:
            addrspace.add_word("two", offset=7)
        yield addrspace


def test_basic(tmp_path, caplog):
    """Address Decoder Basics."""
    decoder = AddrDecoder()
    assert tuple(decoder.addrmap) == ()
    assert tuple(decoder.slaves) == ()
    assert decoder.default_size is None
    assert decoder.is_sub is False

    onemod = Mod(name="one")
    othermod = Mod(name="other")

    one = AddrSlave(name="one", addrdecoder=decoder, ref=onemod)
    decoder.slaves.add(one)

    other = AddrSlave(name="other", addrdecoder=decoder, ref=othermod)
    decoder.slaves.add(other)

    noaddr = AddrSlave(name="noaddr", addrdecoder=decoder)
    decoder.slaves.add(noaddr)

    empty = AddrSlave(name="empty", addrdecoder=decoder)
    decoder.slaves.add(empty)

    nomod = AddrSlave(name="nomod", addrdecoder=decoder, ref=NOREF)
    decoder.slaves.add(nomod)

    assert tuple(decoder.addrmap) == ()
    assert tuple(decoder.slaves) == (one, other, noaddr, empty, nomod)

    addrspaces = (
        one.add_addrrange(size="2k"),
        other.add_addrrange(size="4k"),
        empty.add_addrrange(size="1k"),
        one.add_addrrange(size="2k"),
        nomod.add_addrrange(size="1k"),
        other.add_addrrange(size="4k"),
    )

    assert tuple(decoder.addrmap) == addrspaces
    assert tuple(decoder.slaves) == (one, other, noaddr, empty, nomod)

    resolved = AddrMap.from_addrspaces(decoder.get_addrspaces())

    (tmp_path / "addrmap.md").write_text(decoder.addrmap.get_overview())
    (tmp_path / "resolved.md").write_text(resolved.get_overview())
    assert_refdata(test_basic, tmp_path, caplog=caplog)


def test_addrrange():
    """Test Address Range."""
    decoder = AddrDecoder(default_size="5k")

    one = AddrSlave(name="one", addrdecoder=decoder)
    decoder.slaves.add(one)
    addrspace = one.add_addrrange()
    assert addrspace.size == 5 * 1024

    two = AddrSlave(name="two", addrdecoder=decoder)
    msg = "baseaddr 7168 is not aligned to size 4096"
    with raises(ValueError, match=re.escape(msg)):
        two.add_addrrange(baseaddr=7 * 1024, size=4 * 1024)



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Address Space."""

import re

import ucdp as u
from pytest import raises
from test2ref import assert_refdata

from ucdp_addr import get_addrmap


def test_find_top(tmp_path, testdata_path):
    """Test Find on Example Top Module."""
    mod = u.load("top_lib.top.TopMod").mod
    addrmap = get_addrmap(mod)

    (tmp_path / "overview.md").write_text(addrmap.get_overview())
    assert_refdata(test_find_top, tmp_path)


def test_find_multiple(testdata_path):
    """Multiple."""
    mod = u.load("top_lib.top.SubMod").mod
    msg = (
        "Multiple modules implement 'get_addrspaces':\n"
        "  <top_lib.top.BusMod(inst='sub/u_bus', libname='top_lib', modname='bus')>\n"
        "  <top_lib.top.BusMod(inst='sub/u_otherbus', libname='top_lib', modname='bus')>\n"
        "Implement 'get_addrspaces' on a parent module or choose a different top."
    )
    with raises(ValueError, match=re.escape(msg)):
        get_addrmap(mod)


def test_none(testdata_path):
    """No implementation."""
    mod = u.load("top_lib.top.EmptyMod").mod
    msg = "No module found which implements 'get_addrspaces'"
    with raises(ValueError, match=re.escape(msg)):
        get_addrmap(mod)



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Command Line Interface."""

import click
from ucdp import cli

from .addrmapfinder import get_addrmap

opt_addrdefine = click.option(
    "--addrdefine",
    "-A",
    multiple=True,
    type=str,
    help="Address Defines for Address Map. Environment Variable 'UCDP_ADDRDEFINES'",
    envvar="UCDP_ADDRDEFINES",
)


@click.command(
    help=f"""
Load Data Model and List Address Map.

TOP: Top Module. {cli.PAT_TOPMODREF}. Environment Variable 'UCDP_TOP'
"""
)
@cli.arg_top
@cli.opt_path
@click.option("--full", "-f", is_flag=True)
@opt_addrdefine
@cli.opt_file
@cli.pass_ctx
def lsaddrmap(ctx, top, path, full, addrdefine=None, file=None):
    """Load Data Model and List Address Map."""
    top = cli.load_top(ctx, top, path, quiet=True)
    defines = cli.defines2data(addrdefine)
    addrmap = get_addrmap(top.mod, defines=defines, ref=top.ref)
    print(addrmap.get_overview(minimal=not full), end="", file=file)



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Address Space."""

from test2ref import assert_refdata

from ucdp_addr import AddrMaster, AddrMatrix, AddrSlave


def test_basic(tmp_path, caplog):
    """Basics."""
    matrix = AddrMatrix()
    assert tuple(matrix.masters) == ()
    assert tuple(matrix.slaves) == ()

    (tmp_path / "overview.md").write_text(matrix.get_overview())
    matrix._check_masters_slaves()
    assert_refdata(test_basic, tmp_path, caplog=caplog)


def test_example(tmp_path, caplog):
    """Example."""
    matrix = AddrMatrix()
    mext = AddrMaster(name="ext")
    mdsp = AddrMaster(name="dsp")

    matrix._add_master(mext)
    matrix._add_master(mdsp)

    sram = AddrSlave(name="ram", addrdecoder=matrix)
    speriph = AddrSlave(name="periph", addrdecoder=matrix)
    smisc = AddrSlave(name="misc", addrdecoder=matrix)

    matrix._add_slave(sram, masternames=["ext", "dsp"], baseaddr=0xF000_0000, size=2**16)
    matrix._add_slave(speriph, masternames="dsp", size="32kB")
    matrix._add_slave(smisc, masternames="ext", size="64kB")

    matrix.add_interconnects("dsp", "periph")
    matrix.add_interconnects("ext", "misc")

    assert tuple(matrix.masters) == (mext, mdsp)
    assert tuple(matrix.slaves) == (sram, speriph, smisc)

    (tmp_path / "overview.md").write_text(matrix.get_overview())
    matrix._check_masters_slaves()
    assert_refdata(test_example, tmp_path, caplog=caplog)


def test_corner(tmp_path, caplog):
    """Some Cornercases."""
    matrix = AddrMatrix()
    matrix._add_master(AddrMaster(name="mst"))
    matrix._add_slave(AddrSlave(name="slv", addrdecoder=matrix))
    matrix.add_interconnects("foo", "bar")

    (tmp_path / "overview.md").write_text(matrix.get_overview())
    matrix._check_masters_slaves()
    assert_refdata(test_corner, tmp_path, caplog=caplog)



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Generic Addressed Slave.
"""

import ucdp as u
from icdutil import num

from .addrdecoder import AddrDecoder
from .addrref import AddrRef
from .addrspace import Addrspace


class AddrSlave(u.IdentObject):
    """
    Slave.
    """

    addrdecoder: AddrDecoder = u.Field(repr=False)
    """Demultiplexer Addressing This Slave."""

    ref: AddrRef | None = None
    """Addressed Module."""

    def add_addrrange(self, baseaddr=u.AUTO, size: u.Bytes | None = None) -> "SlaveAddrspace":
        """
        Add Address Range.

        Keyword Args:
            baseaddr: Sub Start Address. Take next free if 'AUTO'.
            size: Address Range Size (i.e. '4k')
            ref: Referenced Object or instance path to it.
        """
        addrdecoder = self.addrdecoder
        addrmap = addrdecoder.addrmap

        # size
        if size is None:
            size = addrdecoder.default_size
        else:
            size = u.Bytesize(size)

        # baseaddr
        if size is not None:
            align = num.calc_next_power_of(size)
            if baseaddr is u.AUTO:
                baseaddr = addrmap.get_free_baseaddr(align)
            if baseaddr != num.align(baseaddr, align=align):
                raise ValueError(f"baseaddr {baseaddr!r} is not aligned to size {align!r}")

        addrspace = SlaveAddrspace(
            name=self.name,
            baseaddr=baseaddr,
            size=size,
            slave=self,
            is_sub=addrdecoder.is_sub,
        )
        addrmap.add(addrspace)
        return addrspace


class SlaveAddrspace(Addrspace):
    """Slave Address Space."""

    slave: AddrSlave



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Address Decoder aka Demultiplexer.
"""

from logging import getLogger

import ucdp as u

from .addrmap import AddrMap
from .addrmapfinder import get_addrspaces
from .addrref import AddrRef
from .addrspaces import Addrspaces, join_addrspaces
from .const import NOREF

LOGGER = getLogger(__name__)


class AddrDecoder(u.Object):
    """
    Address Decoder aka Demultiplexer.

    An address decoder manages the access to a known list of slaves.
    The address map contains the address to slave mapping.
    """

    addrmap: AddrMap = u.Field(default_factory=lambda: AddrMap(unique=True))
    """Address Map With All Slaves."""

    slaves: u.Namespace = u.Field(default_factory=u.Namespace)
    """Namespace With All Slaves."""

    default_size: u.Bytes | None = None
    """Default Size if not given."""

    is_sub: bool = False
    """Just decode address LSBs."""

    def get_addrspaces(self, **kwargs) -> Addrspaces:
        """Address Spaces."""
        for addrspace in self.addrmap:
            ref = addrspace.slave.ref
            if ref is None:
                LOGGER.warning("%r: %r does not reference anything", self, addrspace)
                continue
            if ref is NOREF:
                continue
            ref = self._resolve_ref(ref)
            yield from join_addrspaces(addrspace, get_addrspaces(ref, kwargs))

    def _resolve_ref(self, ref: AddrRef) -> AddrRef:
        return ref



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Access.
"""

from collections.abc import Sequence
from typing import Annotated, Literal

import ucdp as u
from pydantic import (
    BeforeValidator,
    PlainSerializer,
    WithJsonSchema,
)


class ReadOp(u.IdentLightObject):
    """
    Read Operation.

    NEXT = {data}DATA
    """

    data: Literal[None, 0, 1, "~"] = None
    """Operation On Stored Data."""
    once: bool = False
    """Operation is just allowed once."""
    title: str = u.Field(repr=False)
    """Title."""
    descr: str = u.Field(repr=False)
    """Description."""


_R = ReadOp(name="R", title="Read", descr="Read without Modification.")
_RC = ReadOp(name="RC", data=0, title="Read-Clear", descr="Clear on Read.")
_RS = ReadOp(name="RS", data=1, title="Read-Set", descr="Set on Read.")
_RT = ReadOp(name="RT", data="~", title="Read-Toggle", descr="Toggle on Read.")
_RP = ReadOp(name="RP", once=True, title="Read-Protected", descr="Data is hidden after first Read.")


class WriteOp(u.IdentLightObject):
    """
    Write Operation.

    NEXT = {data}DATA {op} {write}WRITE
    """

    data: Literal[None, "", "~"] = None
    """Operation On Stored Data."""
    op: Literal[None, 0, 1, "&", "|"] = None
    """Operation On Stored and Incoming Data."""
    write: Literal[None, "", "~"] = None
    """Operation On Incoming Data."""
    once: bool = False
    """Operation is just allowed once."""
    title: str = u.Field(repr=False)
    """Title."""
    descr: str = u.Field(repr=False)
    """Description."""


_W = WriteOp(name="W", write="", title="Write", descr="Write Data.")
_W0C = WriteOp(name="W0C", data="", op="&", write="", title="Write-Zero-Clear", descr="Clear On Write Zero.")
_W0S = WriteOp(name="W0S", data="", op="|", write="~", title="Write-Zero-Set", descr="Set On Write Zero.")
_W1C = WriteOp(name="W1C", data="", op="&", write="~", title="Write-One-Clear", descr="Clear on Write One.")
_W1S = WriteOp(name="W1S", data="", op="|", write="", title="Write-One-Set", descr="Set on Write One.")
_WL = WriteOp(name="WL", write="", once=True, title="Write Locked", descr="Write Data once and Lock.")


class Access(u.IdentLightObject):
    """Access."""

    read: ReadOp | None = None
    write: WriteOp | None = None

    @property
    def title(self):
        """Title."""
        readtitle = self.read and self.read.title
        writetitle = self.write and self.write.title
        if readtitle and writetitle:
            return f"{readtitle}/{writetitle}"
        return readtitle or writetitle

    @property
    def descr(self):
        """Description."""
        readdescr = self.read and self.read.descr
        writedescr = self.write and self.write.descr
        if readdescr and writedescr:
            return f"{readdescr} {writedescr}"
        return readdescr or writedescr

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


NA = Access(name="NA")
RC = Access(name="RC", read=_RC)
RO = Access(name="RO", read=_R)
RP = Access(name="RP", read=_RP)
RS = Access(name="RS", read=_RS)
RT = Access(name="RT", read=_RT)
RW = Access(name="RW", read=_R, write=_W)
RW0C = Access(name="RW0C", read=_R, write=_W0C)
RW0S = Access(name="RW0S", read=_R, write=_W0S)
RW1C = Access(name="RW1C", read=_R, write=_W1C)
RW1S = Access(name="RW1S", read=_R, write=_W1S)
RWL = Access(name="RWL", read=_R, write=_WL)
W0C = Access(name="W0C", write=_W0C)
W0S = Access(name="W0S", write=_W0S)
W1C = Access(name="W1C", write=_W1C)
W1S = Access(name="W1S", write=_W1S)
WL = Access(name="WL", write=_WL)
WO = Access(name="WO", write=_W)


ACCESSES = u.Namespace(
    (
        NA,
        RC,
        RO,
        RP,
        RS,
        RT,
        RW,
        RW0C,
        RW0S,
        RW1C,
        RW1S,
        RWL,
        W0C,
        W0S,
        W1C,
        W1S,
        WL,
        WO,
    )
)
ACCESSES.lock()


def cast_access(value: str | Access) -> Access:
    """
    Cast Access.

    Usage:

        >>> from ucdp_addr import addrspace
        >>> access = addrspace.cast_access("RO")
        >>> access
        RO
        >>> cast_access(access)
        RO
    """
    if isinstance(value, Access):
        return value
    return ACCESSES[value]


ToAccess = Annotated[
    Access,
    BeforeValidator(lambda x: cast_access(x)),
    PlainSerializer(lambda x: str(x), return_type=str),
    WithJsonSchema({"type": "string"}, mode="serialization"),
]


_COUNTERACCESS = {
    None: RO,
    RO: RW,
    # RC: ,
    # RS: ,
    # RI: ,
    WO: RO,
    # W1C: ,
    # W1S: ,
    RW: RO,
    # RW1C: ,
    # RW1S: ,
    # RCW: ,
    # RCW1C: ,
    # RCW1S: ,
    # RSW: ,
    # RSW1C: ,
    # RSW1S: ,
    # RIW: ,
    # RIW1C: ,
    # RIW1S: ,
}


def get_counteraccess(access: Access) -> Access | None:
    """
    Get Counter Access.

    Usage:

        >>> from ucdp_addr import get_counteraccess, access
        >>> str(get_counteraccess(access.RO))
        'RW'
        >>> str(get_counteraccess(access.RW))
        'RO'
    """
    return _COUNTERACCESS.get(access, None)


def any_read(accesses: Sequence[Access]) -> bool:
    """
    Return `True` if there is any read.

    >>> any_read([NA])
    False
    >>> any_read([RO, RO])
    True
    >>> any_read([RW, RW])
    True
    >>> any_read([RO, RW, WO])
    True
    """
    return any(access.read for access in accesses)


def is_read_repeatable(accesses: Sequence[Access]) -> bool:
    """
    Return `True` if a read has no side-effects.

    >>> is_read_repeatable([NA])
    True
    >>> is_read_repeatable([RO, RO])
    True
    >>> is_read_repeatable([RW, RW])
    True
    >>> is_read_repeatable([RO, RW, WO])
    True
    >>> is_read_repeatable([RO, RW, RC])
    False
    >>> is_read_repeatable([RO, RW, RP])
    False
    """
    return all(not acc.read or not (acc.read.data is not None or acc.read.once) for acc in accesses)


def is_write_repeatable(accesses: Sequence[Access]) -> bool:
    """
    Return `True` if a write has no side-effects.

    >>> is_write_repeatable([NA])
    True
    >>> is_write_repeatable([RO, RO])
    True
    >>> is_write_repeatable([RW, RW])
    True
    >>> is_write_repeatable([RO, RW, WO])
    True
    >>> is_write_repeatable([RO, RW, WL])
    False
    """
    return all(not acc.write or not acc.write.once for acc in accesses)



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Command-Line-Interface."""

import ucdp as u
from click.testing import CliRunner
from pytest import fixture
from test2ref import assert_refdata


@fixture
def runner():
    """Click Runner for Testing."""
    yield CliRunner()


def _run(runner, path, cmd):
    result = runner.invoke(u.cli.ucdp, cmd)
    assert result.exit_code == 0
    (path / "console.md").write_text(result.output)


def test_lsaddrmap(runner, tmp_path, testdata_path):
    """Lsaddrmap Command."""
    _run(runner, tmp_path, ["lsaddrmap", "top_lib.top"])
    assert_refdata(test_lsaddrmap, tmp_path)


def test_lsaddrmap_full(runner, tmp_path, testdata_path):
    """Lsaddrmap Command Full with File."""
    _run(runner, tmp_path, ["lsaddrmap", "top_lib.top", "--full"])
    assert_refdata(test_lsaddrmap_full, tmp_path)


def test_lsaddrmap_full_file(runner, tmp_path, testdata_path):
    """Lsaddrmap Command."""
    filepath = tmp_path / "file.md"
    _run(runner, tmp_path, ["lsaddrmap", "top_lib.top", "--full", "--file", str(filepath)])
    assert_refdata(test_lsaddrmap_full_file, tmp_path)


def test_lsaddrmap_full_define_one(runner, tmp_path, testdata_path):
    """Lsaddrmap Command."""
    _run(runner, tmp_path, ["lsaddrmap", "top_lib.top", "--full", "-A", "one=two"])
    assert_refdata(test_lsaddrmap_full_define_one, tmp_path)


def test_lsaddrmap_full_define_master(runner, tmp_path, testdata_path):
    """Lsaddrmap Command."""
    _run(runner, tmp_path, ["lsaddrmap", "top_lib.top", "--full", "-A", "master=apps"])
    assert_refdata(test_lsaddrmap_full_define_master, tmp_path)



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Address Space."""

import re

import ucdp as u
from pytest import fixture, mark, raises
from test2ref import assert_refdata
from ucdp_glbl.attrs import Attr

from ucdp_addr import DefaultAddrspace, zip_addrspaces
from ucdp_addr.access import ACCESSES, RO, RW
from ucdp_addr.addrspace import (
    Addrspace,
    Field,
    Word,
    Words,
    create_fill_field,
    create_fill_word,
    get_is_const,
    get_is_volatile,
)


@fixture
def word():
    """Example Word."""
    yield Word(name="word", offset=0, width=32)


@fixture
def addrspace():
    """Example Address Space."""
    yield Addrspace(name="name", depth=32)


def test_accesses(tmp_path, capsys):
    """Test Accesses."""
    for access in ACCESSES:
        print(f"{access!s}")
        print(f"    {access.read=!r}")
        print(f"    {access.write=!r}")
        print(f"    {access.title=!r}")
        print(f"    {access.descr=!r}")
    assert_refdata(test_accesses, tmp_path, capsys=capsys)


def test_add_field(addrspace, word):
    """Add Field."""
    field0 = word.add_field("field0", u.UintType(20), "RW")
    assert field0.name == "field0"
    assert field0.type_ == u.UintType(20)
    assert field0.bus == RW
    assert field0.core is None
    assert field0.offset == 0
    assert field0.slice == u.Slice("19:0")
    assert field0.access == "RW/-"

    field1 = word.add_field("field1", u.SintType(6), "RO", core="RO")
    assert field1.name == "field1"
    assert field1.type_ == u.SintType(6)
    assert field1.bus == RO
    assert field1.core == RO
    assert field1.offset == 20
    assert field1.slice == u.Slice("25:20")
    assert field1.access == "RO/RO"

    assert tuple(word.fields) == (field0, field1)

    field2 = word.add_field("field2", u.UintType(2), "RO", align=4)
    assert field2.name == "field2"
    assert field2.type_ == u.UintType(2)
    assert field2.bus == RO
    assert field2.core is None
    assert field2.offset == 28
    assert field2.slice == u.Slice("29:28")
    assert field2.access == "RO/-"

    with raises(ValueError, match=re.escape("Field 'field3' exceeds word width of 32")):
        word.add_field("field3", u.UintType(3), "RO", offset=30)

    field4 = word.add_field("field4", u.UintType(2), "RO", offset=30)
    assert field4.slice == u.Slice("31:30")

    assert tuple(word.fields) == (field0, field1, field2, field4)

    assert addrspace.get_field_hiername(word, field0) == "name.word.field0"
    assert addrspace.get_field_hiername(word, field1) == "name.word.field1"


def test_add_word(addrspace):
    """Add Word."""
    word0 = addrspace.add_word("word0")
    assert word0.name == "word0"
    assert word0.offset == 0
    assert word0.width == 32
    assert word0.depth is None
    assert word0.slice == u.Slice("0")

    word1 = addrspace.add_word("word1", offset=6)
    assert word1.name == "word1"
    assert word1.offset == 6
    assert word1.width == 32
    assert word1.depth is None
    assert word1.slice == u.Slice("6")

    word2 = addrspace.add_word("word2", align=4, depth=2)
    assert word2.name == "word2"
    assert word2.offset == 8
    assert word2.width == 32
    assert word2.depth == 2
    assert word2.slice == u.Slice("9:8")

    with raises(ValueError, match=re.escape("Word 'word3' exceeds address space depth of 32")):
        addrspace.add_word("word3", offset=32)

    word4 = addrspace.add_word("word4", offset=31)
    assert word4.name == "word4"
    assert word4.offset == 31
    assert word4.width == 32
    assert word4.depth is None
    assert word4.slice == u.Slice("31")

    with raises(ValueError, match=re.escape("Word 'word5' has illegal depth of zero.")):
        addrspace.add_word("word5", depth=0)

    assert tuple(addrspace.words) == (word0, word1, word2, word4)

    assert addrspace.get_word_hiername(word0) == "name.word0"
    assert addrspace.get_word_hiername(word1) == "name.word1"

    assert tuple(addrspace.iter()) == ()


def test_addrspace():
    """Address Space."""
    addrspace = Addrspace(name="name", depth=1024)
    assert addrspace.name == "name"
    assert addrspace.words == u.Namespace()
    assert addrspace.width == 32
    assert addrspace.depth == 1024
    assert addrspace.size == u.Bytesize("4 KB")
    assert addrspace.addrwidth == 12
    assert addrspace.baseaddr == u.Hex("0x0")
    assert addrspace.endaddr == u.Hex("0xFFF")
    assert addrspace.nextaddr == u.Hex("0x1000")
    assert addrspace.size_used == u.Bytesize("0 KB")
    assert addrspace.wordsize == 4


def test_addrspace_custom():
    """Address Space."""
    addrspace = Addrspace(name="name", width=64, depth=128, baseaddr=0x10000)
    assert addrspace.name == "name"
    assert addrspace.words == u.Namespace()
    assert addrspace.width == 64
    assert addrspace.depth == 128
    assert addrspace.size == u.Bytesize("1 KB")
    assert addrspace.addrwidth == 10
    assert addrspace.baseaddr == u.Hex("0x10000")
    assert addrspace.endaddr == u.Hex("0x103FF")
    assert addrspace.nextaddr == u.Hex("0x10400")
    assert addrspace.size_used == u.Bytesize("0 KB")


def test_addrspace_size():
    """Address Space with Size."""
    addrspace = Addrspace(name="name", width=64, size="4MB")
    assert addrspace.name == "name"
    assert addrspace.words == u.Namespace()
    assert addrspace.width == 64
    assert addrspace.depth == 4 * 1024 * 1024 // 8
    assert addrspace.size == u.Bytesize("4 MB")
    assert addrspace.addrwidth == 22
    assert addrspace.baseaddr == u.Hex("0x0")
    assert addrspace.endaddr == u.Hex("0x3FFFFF")
    assert addrspace.nextaddr == u.Hex("0x400000")
    assert addrspace.size_used == u.Bytesize("0 KB")
    assert addrspace.wordsize == 8


def test_addrspace_depth_size_both():
    """Address Space with Depth and Size."""
    msg = "'depth' and 'size' are mutually exclusive."
    with raises(ValueError, match=re.escape(msg)):
        Addrspace(depth=128, size=1024)


def test_addrspace_depth_size_missing():
    """Address Space with Depth and Size Missing."""
    msg = "Either 'depth' or 'size' are required."
    with raises(ValueError, match=re.escape(msg)):
        Addrspace()


def test_addrspace_size_used(addrspace):
    """Address Space Size Used."""
    assert addrspace.size_used == u.Bytesize("0 KB")
    addrspace.add_word("one")
    assert addrspace.size_used == u.Bytesize("4 Bytes")
    addrspace.add_word("two", depth=8, align=8)
    assert addrspace.size_used == u.Bytesize("36 Bytes")


def test_lock():
    """Lock Mechanism."""
    addrspace = Addrspace(name="name", depth=1024)
    words = []
    for widx in range(3):
        word = addrspace.add_word(f"word{widx}")
        words.append(word)
        for fidx in range(3):
            word.add_field(f"field{fidx}", u.BitType(), "RO")

    assert addrspace.words.is_locked is False
    assert any(word.fields.is_locked for word in words) is False

    addrspace.lock()

    assert addrspace.words.is_locked is True
    assert all(word.fields.is_locked for word in words) is True


def test_volatile(tmp_path):
    """Test Volatile."""
    overview_file = tmp_path / "overview.md"
    with overview_file.open("w", encoding="utf-8") as file:
        for bus in (None, *ACCESSES):
            for core in (None, *ACCESSES):
                is_volatile = " V" if get_is_volatile(bus, core) else "  "
                is_const = " Const" if get_is_const(bus, core) else " FF"
                busstr = (bus and bus.name) or "-"
                corestr = (core and core.name) or "-"
                file.write(f"{busstr:8s} {corestr:8s}{is_volatile}{is_const}\n")

    assert_refdata(test_volatile, tmp_path)


def test_iter(addrspace):
    """Iteration."""
    a = addrspace.add_word("a")
    a0 = a.add_field("a0", u.BitType(), "RW")

    addrspace.add_word("b")
    c = addrspace.add_word("c")
    c0 = c.add_field("c0", u.BitType(), "RW")
    c1 = c.add_field("c1", u.BitType(), "RW")

    assert tuple(addrspace.iter()) == ((a, (a0,)), (c, (c0, c1)))
    result = ((c, (c0, c1)),)
    assert tuple(addrspace.iter(fieldfilter=lambda field: field.name.startswith("c"))) == result
    result = ((a, (a0,)),)
    assert tuple(addrspace.iter(wordfilter=lambda word: word.name.startswith("a"))) == result


def test_join():
    """Join Operation."""
    f = Addrspace(name="f", baseaddr=0x0, size="4GB", is_sub=False)
    f_s = Addrspace(name="f_s", baseaddr=0x0, size="4GB", is_sub=True)
    m = Addrspace(name="m", baseaddr=0x0000, size="1MB", is_sub=False)
    n = Addrspace(name="n", baseaddr=0xF000_0000, size="1MB", is_sub=False)
    a = Addrspace(name="a", baseaddr=0x2000, size="4KB", is_sub=False)
    a_s = Addrspace(name="a_s", baseaddr=0x2000, size="4KB", is_sub=True)

    assert m.join(a) == a
    assert n.join(a) is None

    assert m.join(a_s) == Addrspace(name="a_s", baseaddr=0x2000, size="4KB", is_sub=False)
    assert n.join(a_s) == Addrspace(name="a_s", baseaddr=0xF000_2000, size="4KB", is_sub=False)

    msg = "does not fit into Addrspace(name='o', "
    with raises(ValueError, match=re.escape(msg)):
        Addrspace(name="o", baseaddr=0xF000_0000, size="2KB", is_sub=False).join(a_s)

    msg = "does not fit into Addrspace(name='a_s', "
    with raises(ValueError, match=re.escape(msg)):
        a_s.join(f_s)

    assert a_s.join(m) == Addrspace(name="a_s", baseaddr=0x2000, depth=1024, is_sub=False)
    assert a_s.join(n) == Addrspace(name="a_s", baseaddr=0xF0002000, depth=1024, is_sub=False)

    assert f.join(f_s) == Addrspace(name="f_s", size="4 GB", is_sub=False)
    assert f_s.join(f) == Addrspace(name="f_s", size="4 GB", is_sub=False)


def test_addrspace_nodefaults():
    """No Defaults."""
    addrspace = Addrspace(name="name", depth=64)
    assert addrspace.bus is None
    assert addrspace.core is None

    word = addrspace.add_word("word")
    assert word.bus is None
    assert word.core is None

    field = word.add_field("field", u.BitType())
    assert field.bus is None
    assert field.core is None


def test_addrspace_defaults():
    """Defaults."""
    addrspace = Addrspace(name="name", bus="RO", core="RW", depth=64)
    assert addrspace.bus == RO
    assert addrspace.core == RW

    word = addrspace.add_word("word")
    assert word.bus == RO
    assert word.core == RW

    field = word.add_field("field", u.BitType())
    assert field.bus == RO
    assert field.core == RW


def test_word_bytewise(addrspace):
    """Test Byte-Wise."""
    word = addrspace.add_word("one", byteoffset=8)
    assert word.offset == 2
    assert word.byteoffset == 8

    word = addrspace.add_word("two", bytealign=8)
    assert word.offset == 4
    assert word.byteoffset == 16

    msg = "'byteoffset' and 'offset' are mutually exclusive"
    with raises(ValueError, match=re.escape(msg)):
        addrspace.add_word("three", byteoffset=8, offset=5)

    msg = "'bytealign' and 'align' are mutually exclusive"
    with raises(ValueError, match=re.escape(msg)):
        addrspace.add_word("three", bytealign=8, align=5)


def test_word_nonbyte():
    """Non-Byte wise."""
    addrspace = Addrspace(name="name", width=12, depth=64)

    msg = "'byteoffset/bytealign' are only available on 'width' multiple of 8 (not 12)"
    with raises(ValueError, match=re.escape(msg)):
        addrspace.add_word("one", byteoffset=4)

    msg = "'byteoffset/bytealign' are only available on 'width' multiple of 8 (not 12)"
    with raises(ValueError, match=re.escape(msg)):
        addrspace.add_word("one", bytealign=4)

    msg = "'byteoffset' is only available on words with 'width' multiple of 8 (not 10)"
    with raises(ValueError, match=re.escape(msg)):
        assert Word(name="word", width=10, offset=2).byteoffset


def test_word_access():
    """Word Access."""
    assert Word(name="word", width=8, offset=2).access == ""

    assert Word(name="word", width=8, offset=2, bus=RO).access == "RO/-"

    assert Word(name="word", width=8, offset=2, core=RO).access == "-/RO"

    assert Word(name="word", width=8, offset=2, bus=RW, core=RO).access == "RW/RO"


def test_addrspace_access():
    """Addressspace Access."""
    assert Addrspace(name="word", width=8, depth=2).access == ""

    assert Addrspace(name="word", width=8, depth=2, bus=RO).access == "RO/-"

    assert Addrspace(name="word", width=8, depth=2, core=RO).access == "-/RO"

    assert Addrspace(name="word", width=8, depth=2, bus=RW, core=RO).access == "RW/RO"


@fixture
def sparse_addrspace():
    """Sparse Address Space."""
    addrspace = Addrspace(name="name", depth=32)

    addrspace.add_word("empty", offset=2)

    word = addrspace.add_word("word0", offset=10)
    word.add_field("field0", u.UintType(2))

    word = addrspace.add_word("word1")
    word.add_field("field1", u.UintType(32))

    word = addrspace.add_word("word2", offset=20)
    word.add_field("field0", u.UintType(2), offset=8)
    word.add_field("field1", u.UintType(4), offset=16)

    yield addrspace


def _dump_addrspace(iterable, tmp_path, name="dump"):
    with (tmp_path / f"{name}.txt").open("w") as file:
        for word, fields in iterable:
            file.write(f"{word}\n")
            for field in fields:
                file.write(f"  {field}\n")


def _create_word(addrspace, idx, offset, depth):
    return Word(name=f"fillword{idx}", offset=offset, depth=depth, width=addrspace.width)


def _create_field(word, idx, offset, width):
    return Field(name=f"fillfield{idx}", type_=u.UintType(width), offset=offset)


def test_addrspace_iter(sparse_addrspace, tmp_path):
    """Iterate."""
    _dump_addrspace(sparse_addrspace.iter(), tmp_path)
    assert_refdata(test_addrspace_iter, tmp_path)


def test_addrspace_iter_fill_word(sparse_addrspace, tmp_path):
    """Iterate with word filling."""
    _dump_addrspace(sparse_addrspace.iter(fill_word=_create_word), tmp_path, name="dump")
    _dump_addrspace(sparse_addrspace.iter(fill_word=True), tmp_path, name="dump-true")
    _dump_addrspace(
        sparse_addrspace.iter(fill_word=True, wordfilter=lambda word: word.name != "word0"),
        tmp_path,
        name="dump-true-filter",
    )
    _dump_addrspace(sparse_addrspace.iter(fill_word=_create_word, fill_word_end=True), tmp_path, name="dump_end")
    assert_refdata(test_addrspace_iter_fill_word, tmp_path)


def test_addrspace_iter_fill_field(sparse_addrspace, tmp_path):
    """Iterate with field filling."""
    _dump_addrspace(sparse_addrspace.iter(fill_field=_create_field), tmp_path, name="dump")
    _dump_addrspace(sparse_addrspace.iter(fill_field=True), tmp_path, name="dump-true")
    _dump_addrspace(
        sparse_addrspace.iter(fill_field=True, fieldfilter=lambda field: field.name != "field1"),
        tmp_path,
        name="dump-true-filter",
    )
    _dump_addrspace(sparse_addrspace.iter(fill_field=_create_field, fill_field_end=True), tmp_path, name="dump_end")
    assert_refdata(test_addrspace_iter_fill_field, tmp_path)


def test_addrspace_iter_fill_word_fill_field(sparse_addrspace, tmp_path):
    """Iterate with filling."""
    _dump_addrspace(sparse_addrspace.iter(fill_word=_create_word, fill_field=_create_field), tmp_path, name="dump_0_0")
    _dump_addrspace(
        sparse_addrspace.iter(fill_word=_create_word, fill_field=_create_field, fill_word_end=True),
        tmp_path,
        name="dump_1_0",
    )
    _dump_addrspace(
        sparse_addrspace.iter(fill_word=_create_word, fill_field=_create_field, fill_field_end=True),
        tmp_path,
        name="dump_0_1",
    )
    _dump_addrspace(
        sparse_addrspace.iter(
            fill_word=_create_word, fill_field=_create_field, fill_word_end=True, fill_field_end=True
        ),
        tmp_path,
        name="dump_1_1",
    )
    assert_refdata(test_addrspace_iter_fill_word_fill_field, tmp_path)


def test_addrspace_iter_fill_defaults(sparse_addrspace, tmp_path):
    """Iterate with filling."""
    _dump_addrspace(
        sparse_addrspace.iter(fill_word=create_fill_word, fill_field=create_fill_field),
        tmp_path,
    )
    assert_refdata(test_addrspace_iter_fill_defaults, tmp_path)


def test_zip():
    """Zip."""
    a0 = Addrspace(name="a0", baseaddr=0x0000, size=0x1000)
    a1 = Addrspace(name="a1", baseaddr=0x1000, size=0x1000)
    b0 = Addrspace(name="b0", baseaddr=0x0000, size=0x1000)
    b1 = Addrspace(name="b1", baseaddr=0x1000, size=0x800)
    b2 = Addrspace(name="b2", baseaddr=0x1800, size=0x800)

    assert tuple(zip_addrspaces((), (b0,))) == ()
    assert tuple(zip_addrspaces((a0, a1), (b0, b1))) == ((a0, b0), (a1, b1))
    assert tuple(zip_addrspaces((a0,), (b0,))) == ((a0, b0),)
    assert tuple(zip_addrspaces((a0, a1), (b0, b1, b2))) == ((a0, b0), (a1, b1), (a1, b2))


def test_ident():
    """Ident."""
    addrspace = Addrspace(name="name", width=32, depth=32)
    with raises(u.ValidationError):
        addrspace.add_word("foo ")

    word = addrspace.add_word("foo")
    with raises(u.ValidationError):
        word.add_field("foo ", type_=u.UintType(2))


def test_default_addrspace():
    """Default Address Space."""
    default = DefaultAddrspace(size="4kb", attrs="one")
    assert default.name == ""
    assert default.size == 4096
    assert default.attrs == (Attr("one"),)


def test_add_words(tmp_path):
    """Iterate with field filling."""
    addrspace = Addrspace(name="module", width=32, depth=128)

    word = addrspace.add_words("a")
    word.add_field("a", u.UintType(14), "RW")
    word.add_field("b", u.UintType(14), "RW")
    word.add_field("c", u.UintType(14), "RW")
    word.add_field("d", u.UintType(14), "RW")

    word = addrspace.add_words("b", align=16)
    word.add_field("a", u.UintType(14), "RW")
    word.add_field("b", u.UintType(18), "RW")
    word.add_field("c", u.UintType(10), "RW")
    word.add_field("d", u.UintType(23), "RW")

    _dump_addrspace(addrspace.iter(fill=True), tmp_path)
    assert_refdata(test_add_words, tmp_path)


def _mynaming(value: int) -> str:
    return chr(65 + value)


@mark.parametrize("naming", ("dec", "alpha", _mynaming))
def test_add_words_naming(tmp_path, naming):
    """Iterate with field filling."""
    addrspace = Addrspace(name="module", width=32, depth=128, add_words_naming=naming)

    word = addrspace.add_words("a")
    word.add_field("a", u.UintType(14), "RW")
    word.add_field("b", u.UintType(14), "RW")
    word.add_field("c", u.UintType(14), "RW")
    word.add_field("d", u.UintType(14), "RW")

    for name, wordnaming in (("b", "dec"), ("c", "alpha"), ("d", _mynaming)):
        word = addrspace.add_words(name, align=16, naming=wordnaming)
        word.add_field("a", u.UintType(14), "RW")
        word.add_field("b", u.UintType(18), "RW")
        word.add_field("c", u.UintType(10), "RW")
        word.add_field("d", u.UintType(23), "RW")

    flavor = "func" if naming is _mynaming else naming
    _dump_addrspace(addrspace.iter(), tmp_path)
    assert_refdata(test_add_words_naming, tmp_path, flavor=flavor)


class MyField(Field):
    """Example Field Implementation."""

    basename: str


class MyWord(Word):
    """Example Word Implementation."""

    def _create_field(self, **kwargs) -> Field:
        name = kwargs.pop("name")
        basename = kwargs.pop("basename", None) or f"{self.name}_{name}"
        return MyField(name=name, basename=basename, **kwargs)


class MyWords(Words):
    """Example Words Implementation."""

    def _add_field(self, name: str, type_: u.BaseScalarType, *args, **kwargs):
        basename = kwargs.pop("basename", None) or f"{self.name}_{name}"
        self.word.add_field(name, type_, *args, basename=basename, **kwargs)


class MyAddrspace(Addrspace):
    """Example Addrspace Implementation."""

    def _create_word(self, **kwargs) -> Word:
        return MyWord(**kwargs)

    def _create_words(self, **kwargs) -> Words:
        return MyWords.create(**kwargs)


def test_add_words_custom(tmp_path):
    """Iterate with field filling."""
    addrspace = MyAddrspace(name="module", width=32, depth=128)

    word = addrspace.add_words("a")
    word.add_field("a", u.UintType(14), "RW")
    word.add_field("b", u.UintType(14), "RW", basename="base")
    word.add_field("c", u.UintType(14), "RW")
    word.add_field("d", u.UintType(14), "RW")

    word = addrspace.add_words("b", align=16)
    word.add_field("a", u.UintType(14), "RW")
    word.add_field("b", u.UintType(18), "RW", basename="base")
    word.add_field("c", u.UintType(10), "RW")
    word.add_field("d", u.UintType(23), "RW")

    _dump_addrspace(addrspace.iter(), tmp_path)
    assert_refdata(test_add_words_custom, tmp_path)



"""Module, Word and Field Reference."""

from typing import TypeAlias

import ucdp as u

from .addrrange import AddrRange
from .addrspace import Addrspace, Field, Word


class AddrMapRef(u.Object):
    """Address Map Reference."""

    addrrange: AddrRange
    """Address Range."""

    addrspace: Addrspace | None = None
    """Address Space."""

    word: Word | None = None
    """Word."""

    field: Field | None = None
    """Field."""

    def __str__(self) -> str:
        if self.addrspace:
            result = f"{self.addrspace.name}"
            if self.word:
                result = f"{result}.{self.word.name}"
                if self.field:
                    result = f"{result}.{self.field.name}"
            return result
        return f"{self.addrrange.baseaddr}"

    @staticmethod
    def create(
        addrspace: Addrspace, word: Word | None = None, field: Field | None = None, addrrange: AddrRange | None = None
    ) -> "AddrMapRef":
        """
        Create Addrspace with Proper AddrRange.

        Args:
            addrspace: Address Space

        Keyword Args:
            word: Word
            field: Field (requires word as well)
            addrrange: Address Range
        """
        if addrrange is None:
            if word:
                addrrange = AddrRange(
                    baseaddr=addrspace.baseaddr + word.byteoffset,
                    width=word.width,
                    depth=word.depth or 1,
                )
            else:
                addrrange = AddrRange(baseaddr=addrspace.baseaddr, width=addrspace.width, depth=addrspace.depth)
        return AddrMapRef(addrrange=addrrange, addrspace=addrspace, word=word, field=field)


ToAddrMapRef: TypeAlias = AddrMapRef | AddrRange | Addrspace | str | int
"""Unresolved Address Map Reference."""

RawAddrMapRef = ToAddrMapRef
"""Obsolete Alias."""



"""Module, Word and Field Resolver."""

import re
from collections.abc import Iterator

from matchor import match

from .addrmap import AddrMap
from .addrmapref import AddrMapRef, ToAddrMapRef
from .addrrange import AddrRange
from .addrspace import Addrspace, Word

_RE_RESOLVE = re.compile(
    r"\A(?P<apat>[A-Za-z\*\?][A-Za-z0-9_\*\?]*)"
    r"(\.(?P<wpat>[A-Za-z\*\?][A-Za-z0-9_\*\?]*)"
    r"(\.(?P<fpat>[A-Za-z][A-Za-z0-9_\*\?]*))?)?\Z"
)


def resolve(addrmap: AddrMap, item: ToAddrMapRef) -> AddrMapRef:
    """
    Retrieve AddrSpace, Word and Field referenced by `item` from `addrmap`.

    `item` may be a string pattern with word and field name separated by '.' according to this scheme:
    'addrspace[.word[.field]]'. Wildcards '*' and '?' are supported. The first match is returned.

    `item` may be an address (integer). On `check=True` the address is checked to be covered by a valid
    addrspace and word, if the addrspace contains words.

    `item` may be already a address map reference. On `check=True` the address is checked to be covered by a valid
    addrspace and word, if the addrspace contains words.

    Example:
    >>> import ucdp as u
    >>> from ucdp_addr import Addrspace, Word, Field, AddrMap
    >>> addrmap = AddrMap()
    >>> for aname in ("uart", "spi", "owi"):
    ...     addrspace = Addrspace(name=aname, width=32, depth=128)
    ...     addrmap.add(addrspace)
    ...     word = addrspace.add_word("ctrl")
    ...     field = word.add_field("ena", u.BitType(), "RW")
    ...     word = addrspace.add_word("stat")
    ...     field = word.add_field("bsy", u.BitType(), "RO")

    By string:

    >>> resolve(addrmap, "uart")
    AddrMapRef(..., addrspace=Addrspace(name='uart', ...))
    >>> resolve(addrmap, "uart.ctrl")
    AddrMapRef(..., addrspace=Addrspace(name='uart', ...), word=Word(name='ctrl', ...))
    >>> resolve(addrmap, "uart.ctrl.ena")
    AddrMapRef(..., addrspace=Addrspace(name='uart', ...), word=Word(name='ctrl', ...), field=Field(name='ena', ...))

    By reference:

    >>> ref = resolve(addrmap, "uart.ctrl.ena")
    >>> ref
    AddrMapRef(..., addrspace=Addrspace(name='uart', ...), word=Word(name='ctrl', ...), field=Field(name='ena', ...))
    >>> resolve(addrmap, ref)
    AddrMapRef(..., addrspace=Addrspace(name='uart', ...), word=Word(name='ctrl', ...), field=Field(name='ena', ...))

    By address space:

    >>> resolve(addrmap, Addrspace(baseaddr=0, width=32, depth=1))
    AddrMapRef(..., addrspace=Addrspace(name='uart', size=Bytesize('4 bytes')))

    By address range:

    >>> resolve(addrmap, AddrRange(baseaddr=0, width=32, depth=1))
    AddrMapRef(addrrange=AddrRange(size=Bytesize('4 bytes')))

    By address:

    >>> resolve(addrmap, 8)
    AddrMapRef(addrrange=AddrRange(baseaddr=Hex('0x8'), size=Bytesize('4 bytes')))

    Errors:

    >>> resolve(addrmap, "uart.missing")
    Traceback (most recent call last):
      ...
    ValueError: 'uart.missing' does not exists
    >>> resolve(addrmap, "uart:ctrl")
    Traceback (most recent call last):
      ...
    ValueError: uart:ctrl does not match pattern 'addrspace[.word[.field]]'
    >>> resolve(addrmap, 5.0)
    Traceback (most recent call last):
    ...
    TypeError: 5.0
    >>> resolve(addrmap, Addrspace(baseaddr=0x1000, width=32, depth=1))
    Traceback (most recent call last):
      ...
    ValueError: Addrspace(baseaddr=Hex('0x1000'), size=Bytesize('4 bytes')) does not exists
    """
    try:
        return next(iter(resolves(addrmap, item)))
    except StopIteration:
        raise ValueError(f"{item!r} does not exists") from None


def resolves(addrmap: AddrMap, item: ToAddrMapRef) -> Iterator[AddrMapRef]:
    """
    Same as `resolve` but returns all matches.

    >>> import ucdp as u
    >>> from ucdp_addr import Addrspace, Word, Field, AddrMap
    >>> addrmap = AddrMap()
    >>> for aname in ("uart", "spi", "owi"):
    ...     addrspace = Addrspace(name=aname, width=32, depth=128)
    ...     addrmap.add(addrspace)
    ...     word = addrspace.add_word("ctrl")
    ...     field = word.add_field("ena", u.BitType(), "RW")
    ...     word = addrspace.add_word("stat")
    ...     field = word.add_field("bsy", u.BitType(), "RO")

    >>> for item in resolves(addrmap, "*.ctrl.ena"): print(item)
    uart.ctrl.ena
    spi.ctrl.ena
    owi.ctrl.ena
    """
    if isinstance(item, str):
        # 'addrspace[.word[.field]]' reference
        yield from _resolve_pat(addrmap, item)

    elif isinstance(item, AddrMapRef):
        yield item

    elif isinstance(item, Addrspace):
        yield addrmap.get(item)

    elif isinstance(item, AddrRange):
        yield AddrMapRef(addrrange=item)

    elif isinstance(item, int):
        yield AddrMapRef(addrrange=AddrRange(baseaddr=item, size=4))

    else:
        raise TypeError(item)


def _resolve_pat(addrmap: AddrMap, pat: str) -> Iterator[AddrMapRef]:
    mat = _RE_RESOLVE.match(pat)
    if mat:
        apat, wpat, fpat = mat.group("apat", "wpat", "fpat")
        addrspaces = [addrspace for addrspace in addrmap.iter() if match(addrspace.name, apat)]
        if wpat:
            addrspace_words: list[tuple[Addrspace, Word]] = []
            for addrspace in addrspaces:
                addrspace_words.extend((addrspace, word) for word in addrspace.words if match(word.name, wpat))

            if fpat:
                # 'addrspace.word.field'
                for addrspace, word in addrspace_words:
                    for field in word.fields:
                        if match(field.name, fpat):
                            yield AddrMapRef.create(addrspace=addrspace, word=word, field=field)
            else:
                # 'addrspace.word'
                for addrspace, word in addrspace_words:
                    yield AddrMapRef.create(addrspace=addrspace, word=word)
        else:
            # 'addrspace'
            for addrspace in addrspaces:
                yield AddrMapRef.create(addrspace=addrspace)
    else:
        raise ValueError(f"{pat} does not match pattern 'addrspace[.word[.field]]'")



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 1
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. @IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Address Map Finder.
"""

from collections.abc import Callable

import ucdp as u
from ucdp_glbl.finder import find

from .addrmap import AddrMap, Defines
from .addrspaces import Addrspaces

GetAttrspacesFunc = Callable[[], Addrspaces]


def get_addrmap(
    mod: u.BaseMod, defines: Defines | None = None, unique: bool = False, ref: u.TopModRef | None = None
) -> AddrMap:
    """Search Address Spaces and Create Address Map."""
    defines = defines or {}
    addrmap = AddrMap(unique=unique, defines=defines, ref=ref)
    for addrspace in get_addrspaces(mod, defines=defines):
        addrmap.add(addrspace)
    return addrmap


def get_addrspaces(mod: u.BaseMod, defines: Defines | None = None) -> Addrspaces:
    """Search Address Spaces."""
    func = _find_get_addrspaces([mod])
    defines = defines or {}
    return func(**defines)


def _find_get_addrspaces(mods: list[u.BaseMod]) -> GetAttrspacesFunc:
    return find(mods, "get_addrspaces")[1]



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Address Space.
"""

from collections import defaultdict
from collections.abc import Callable, Iterator
from string import ascii_lowercase
from typing import Any, Literal, Optional, TypeAlias

import pydantic as pyd
import ucdp as u
from humannum import bytesize_, int_
from icdutil import num
from ucdp_glbl.attrs import CastableAttrs

from .access import Access, ToAccess, cast_access
from .addrrange import AddrRange

NamingScheme: TypeAlias = Literal["dec", "alpha"] | Callable[[int], str]

_ALPHA_PER_DIGIT = len(ascii_lowercase)


class FullError(ValueError):
    """Full."""


class Field(u.IdentLightObject):
    """Field."""

    type_: u.BaseScalarType
    """Type."""
    bus: Access | None = None
    """Bus Access."""
    core: Access | None = None
    """Core Access."""
    offset: int | u.Expr
    """Rightmost Bit Position."""
    is_volatile: bool = False
    """Volatile."""
    doc: u.Doc = u.Doc()
    """Documentation."""
    attrs: CastableAttrs = ()
    """Attributes."""

    @property
    def slice(self) -> u.Slice:
        """Slice with Word."""
        return u.Slice(width=self.type_.width, right=self.offset)

    @property
    def is_const(self) -> bool:
        """Field is Constant."""
        return get_is_const(self.bus, self.core)

    @property
    def access(self) -> str:
        """Access."""
        bus = (self.bus and self.bus.name) or "-"
        core = (self.core and self.core.name) or "-"
        return f"{bus}/{core}"


FieldFilter = Callable[[Field], bool]


class Word(u.IdentObject):
    """Word."""

    fields: u.Namespace = u.Field(default_factory=u.Namespace, init=False, repr=False)
    """Fields within Word."""
    offset: int | u.Expr
    """Rightmost Word Position."""
    width: int
    """Width in Bits."""
    depth: int | u.Expr | None = None
    """Number of words."""
    doc: u.Doc = u.Doc()
    """Documentation"""
    attrs: CastableAttrs = ()
    """Attributes."""

    bus: Access | None = None
    core: Access | None = None
    is_volatile: bool | None = None

    def add_field(
        self,
        name: str,
        type_: u.BaseScalarType,
        bus: Access | None = None,
        core: Access | None = None,
        offset: int | u.Expr | None = None,
        align: int | u.Expr | None = None,
        is_volatile: bool | None = None,
        title: str | None = None,
        descr: str | None = None,
        comment: str | None = None,
        **kwargs,
    ) -> Field:
        """Add field."""
        if bus is not None:
            bus = cast_access(bus)
        else:
            bus = self.bus
        if core is not None:
            core = cast_access(core)
        else:
            core = self.core
        if self.fields:
            free = tuple(self.fields)[-1].slice.left + 1
        else:
            free = 0
        offset = num.align(free, offset=offset, align=align)
        doc = u.doc_from_type(type_, title=title, descr=descr, comment=comment)
        if is_volatile is None:
            is_volatile = self.is_volatile
        if is_volatile is None:
            is_volatile = get_is_volatile(bus, core)
        field = self._create_field(
            name=name,
            type_=type_,
            bus=bus,
            core=core,
            offset=offset,
            is_volatile=is_volatile,
            doc=doc,
            **kwargs,
        )
        if field.slice.left >= self.width:
            raise FullError(f"Field {field.name!r} exceeds word width of {self.width}")
        self.fields.add(field)
        return field

    def _create_field(self, **kwargs) -> Field:
        return Field(**kwargs)

    @property
    def slice(self) -> u.Slice:
        """Slice with Address Space."""
        if self.depth is None:
            return u.Slice(left=self.offset, right=self.offset)
        return u.Slice(width=self.depth, right=self.offset)

    def lock(self):
        """Lock For Modification."""
        self.fields.lock()

    @property
    def wordsize(self) -> float:
        """Word Size in Bytes."""
        return self.width * (self.depth or 1) / 8

    @property
    def byteoffset(self) -> u.Hex:
        """Offset in Bytes, if word width is multiple of 8."""
        if (self.width % 8) != 0:
            raise ValueError(f"'byteoffset' is only available on words with 'width' multiple of 8 (not {self.width})")
        return self.offset * self.width // 8

    @property
    def access(self) -> str:
        """Access."""
        if self.bus or self.core:
            bus = (self.bus and self.bus.name) or "-"
            core = (self.core and self.core.name) or "-"
            return f"{bus}/{core}"
        return ""


def _bytealign(width, free, offset=None, align=None, byteoffset=None, bytealign=None):
    if byteoffset is None and bytealign is None:
        pass
    elif (width % 8) == 0:
        bytesperword = width // 8
        if byteoffset is not None:
            if offset is not None:
                raise ValueError("'byteoffset' and 'offset' are mutually exclusive")
            offset = byteoffset // bytesperword
        if bytealign is not None:
            if align is not None:
                raise ValueError("'bytealign' and 'align' are mutually exclusive")
            align = bytealign // bytesperword
    else:
        raise ValueError(f"'byteoffset/bytealign' are only available on 'width' multiple of 8 (not {width})")
    return num.align(free, offset=offset, align=align)


WordFilter = Callable[[Word], bool]
WordFields = tuple[Word, tuple[Field, ...]]

FillWordFactory = Callable[["Addrspace", int, int, int], Word]
FillFieldFactory = Callable[[Word, int, int, int], Field]


class Words(u.Object):
    """Multiple Related Words."""

    model_config = pyd.ConfigDict(
        frozen=False,
    )

    name: str
    addrspace: "Addrspace"
    word_kwargs: dict[str, Any]
    naming: NamingScheme = "dec"

    idx: int
    word: Word

    @classmethod
    def create(
        cls, name: str, addrspace: "Addrspace", word_kwargs: dict[str, Any], naming: NamingScheme = "dec", **kwargs
    ) -> "Words":
        """Create Helper for Set of Words."""
        idx, word = cls._create_word(name, addrspace, word_kwargs, naming, **kwargs)
        return cls(name=name, addrspace=addrspace, word_kwargs=word_kwargs, idx=idx, word=word, naming=naming)

    @staticmethod
    def _create_word(
        name: str, addrspace: "Addrspace", word_kwargs: dict[str, Any], naming: NamingScheme, idx: int = 0, **kwargs
    ) -> tuple[int, Word]:
        if naming == "dec":
            suffix = str(idx)
        elif naming == "alpha":
            suffix = name_alpha(idx)
        else:
            suffix = naming(idx)
        word = addrspace.add_word(f"{name}{suffix}", **word_kwargs, **kwargs)
        return idx + 1, word

    def _add_field(self, *args, **kwargs):
        self.word.add_field(*args, **kwargs)

    def next(self):
        """Start a new Word."""
        self.idx, self.word = self._create_word(self.name, self.addrspace, self.word_kwargs, self.naming, idx=self.idx)

    def add_field(self, *args, **kwargs):
        """Add Field to Current Word or start a new one."""
        try:
            self._add_field(*args, **kwargs)
        except FullError:
            self.next()
            self._add_field(*args, **kwargs)


class Addrspace(AddrRange, u.IdentObject):
    """Address Space."""

    name: str = ""
    """Name."""
    is_sub: bool = True
    """Address Decoder Just Compares `addrwidth` LSBs."""
    words: u.Namespace = u.Field(default_factory=u.Namespace, repr=False)
    """Words within Address Space."""
    attrs: CastableAttrs = ()
    """Attributes."""

    add_words_naming: NamingScheme = "dec"
    """Naming Scheme for words created by `add_words`."""

    bus: ToAccess | None = None
    core: ToAccess | None = None
    is_volatile: bool | None = None

    @property
    def size_used(self) -> u.Bytes:
        """Number of Bytes Used."""
        return bytesize_(int(sum(word.wordsize for word in self.words)))

    @property
    def free_offset(self) -> int:
        """Free Offset."""
        if self.words:
            return tuple(self.words)[-1].slice.left + 1
        return 0

    @property
    def info(self) -> str:
        """Info."""
        return str(self)

    def __str__(self) -> str:
        baseaddr = f"+{self.baseaddr}" if self.is_sub else f"{self.baseaddr}"
        return f"{self.name} {baseaddr} {self.depth}x{self.width}"

    @property
    def base(self) -> str:
        """Base."""
        if self.is_sub:
            return f"+{self.baseaddr}"
        return f"{self.baseaddr}"

    @property
    def access(self) -> str:
        """Access."""
        if self.bus or self.core:
            bus = (self.bus and self.bus.name) or "-"
            core = (self.core and self.core.name) or "-"
            return f"{bus}/{core}"
        return ""

    def add_word(
        self,
        name: str,
        offset: int | u.Expr | None = None,
        align: int | u.Expr | None = None,
        byteoffset: int | u.Expr | None = None,
        bytealign: int | u.Expr | None = None,
        depth: int | u.Expr | None = None,
        bus: Access | None = None,
        core: Access | None = None,
        is_volatile: bool | None = None,
        title: str | None = None,
        descr: str | None = None,
        comment: str | None = None,
        **kwargs,
    ) -> Word:
        """Add Word."""
        free = self.free_offset
        offset = _bytealign(self.width, free, offset=offset, align=align, byteoffset=byteoffset, bytealign=bytealign)
        doc = u.Doc(title=title, descr=descr, comment=comment)
        if bus is None:
            bus = self.bus
        if core is None:
            core = self.core
        if is_volatile is None:
            is_volatile = self.is_volatile
        word = self._create_word(
            name=name,
            offset=offset,
            width=self.width,
            depth=depth,
            doc=doc,
            bus=bus,
            core=core,
            is_volatile=is_volatile,
            **kwargs,
        )
        if word.slice.left >= self.depth:
            raise FullError(f"Word {word.name!r} exceeds address space depth of {self.depth}")
        if word.depth == 0:
            raise ValueError(f"Word {word.name!r} has illegal depth of zero.")
        self.words.add(word)
        return word

    def add_words(
        self,
        name: str,
        offset: int | u.Expr | None = None,
        align: int | u.Expr | None = None,
        byteoffset: int | u.Expr | None = None,
        bytealign: int | u.Expr | None = None,
        depth: int | u.Expr | None = None,
        naming: NamingScheme | None = None,
        **kwargs,
    ) -> Words:
        """Add Word."""
        if depth is not None:
            raise ValueError("'depth' is not supported on add_words()")
        return self._create_words(
            name=name,
            addrspace=self,
            word_kwargs=kwargs,
            offset=offset,
            align=align,
            byteoffset=byteoffset,
            bytealign=bytealign,
            naming=naming or self.add_words_naming,
        )

    def _create_word(self, **kwargs) -> Word:
        return Word(**kwargs)

    def _create_words(self, **kwargs) -> Words:
        return Words.create(**kwargs)

    def lock(self):
        """Lock For Modification."""
        for word in self.words:
            word.lock()
        self.words.lock()

    def get_word_hiername(self, word: Word) -> str:
        """Get Hierarchical Word Name."""
        return f"{self.name}.{word.name}"

    def get_field_hiername(self, word: Word, field: Field) -> str:
        """Get Hierarchical Field Name."""
        return f"{self.name}.{word.name}.{field.name}"

    def iter(  # noqa: C901
        self,
        wordfilter: WordFilter | None = None,
        fieldfilter: FieldFilter | None = None,
        fill: bool | None = None,
        fill_word: FillWordFactory | bool | None = None,
        fill_field: FillFieldFactory | bool | None = None,
        fill_word_end: bool | None = None,
        fill_field_end: bool | None = None,
    ) -> Iterator[WordFields]:
        """Iterate over words and their fields."""
        if fill is not None:
            fill_word = fill if fill_word is None else fill_word
            fill_field = fill if fill_field is None else fill_field
            fill_word_end = fill if fill_word_end is None else fill_word_end
            fill_field_end = fill if fill_field_end is None else fill_field_end

        def no_wordfilter(_: Word) -> bool:
            return True

        wordfilter = wordfilter or no_wordfilter

        def no_fieldfilter(_: Word) -> bool:
            return True

        fieldfilter = fieldfilter or no_fieldfilter

        if fill_word is True:
            fill_word = create_fill_word

        if fill_field is True:
            fill_field = create_fill_field

        def create_word(counter, offset, size) -> WordFields:
            idx = counter[None]
            fword = fill_word(self, idx, offset, size)  # type: ignore[operator]
            counter[None] = idx + 1
            if fill_field and fill_field_end:
                return fword, (fill_field(fword, 0, 0, word.width),)
            return fword, ()

        def fill_fields(word, fields) -> Iterator:
            offset = 0
            idx = 0
            for field in fields:
                if not fieldfilter(field):
                    continue

                # fill before
                if field.offset > offset:
                    yield fill_field(word, idx, offset, field.offset - offset)
                    idx += 1

                yield field

                offset = field.slice.nxt

            if fill_field_end:
                if word.width > offset:
                    yield fill_field(word, idx, offset, word.width - offset)

        counter: dict[None, int] = defaultdict(int)
        offset = 0
        for word in self.words:
            if not wordfilter(word):
                continue

            # fill before
            if fill_word and word.offset > offset:
                yield create_word(counter, offset, word.offset - offset)
                offset = word.offset + (word.depth or 1)

            if fill_field:
                fields = tuple(fill_fields(word, word.fields))
            else:
                fields = tuple(field for field in word.fields if fieldfilter(field))

            if fields:
                yield word, fields
                offset = word.offset + (word.depth or 1)

        if fill_word and fill_word_end:
            yield create_word(counter, offset, self.depth - offset)

    def is_overlapping(self, other: "Addrspace") -> bool:
        """
        Determine both Address Spaces Overlap.

            >>> one = Addrspace(name='one', baseaddr=0x2000, size='4kB')
            >>> two = Addrspace(name='two', baseaddr=0x3000, size='4kB')
            >>> three = Addrspace(name='three', baseaddr=0x2000, size='5kB')
            >>> one.is_overlapping(two)
            False
            >>> one.is_overlapping(three)
            True
            >>> three.is_overlapping(one)
            True
            >>> three.is_overlapping(two)
            True
        """
        if self.baseaddr < other.baseaddr:
            return self.endaddr >= other.baseaddr
        return self.baseaddr <= other.endaddr

    def get_intersect(self, other: "Addrspace") -> Optional["Addrspace"]:
        """
        Get Intersection.

            >>> one = Addrspace(name='one', baseaddr=0x2000, size='4kB')
            >>> two = Addrspace(name='two', baseaddr=0x3000, size='4kB')
            >>> three = Addrspace(name='three', baseaddr=0x2000, size='5kB')
            >>> one.get_intersect(two)
            >>> one.get_intersect(three)
            Addrspace(name='one', baseaddr=Hex('0x2000'), size=Bytesize('4 KB'))
            >>> three.get_intersect(one)
            Addrspace(name='three', baseaddr=Hex('0x2000'), size=Bytesize('4 KB'))
            >>> three.get_intersect(two)
            Addrspace(name='three', baseaddr=Hex('0x3000'), size=Bytesize('1 KB'))
        """
        baseaddr = max(self.baseaddr, other.baseaddr)
        endaddr = min(self.endaddr, other.endaddr)
        size = endaddr - baseaddr + 1
        if size <= 0:
            return None
        words = u.Namespace()
        # TODO: add words
        return self.new(baseaddr=baseaddr, size=size, depth=None, words=words)

    def join(self, other: "Addrspace") -> Optional["Addrspace"]:
        """
        Join if Possible.

        TODO: doc
        """
        if other.is_sub:
            return self._join(self, other)
        if self.is_sub:
            return self._join(other, self)
        return other.get_intersect(self)

    @staticmethod
    def _join(one: "Addrspace", sub: "Addrspace") -> Optional["Addrspace"]:
        baseaddr = one.baseaddr + sub.baseaddr
        endaddr = one.baseaddr + sub.endaddr
        if endaddr > one.endaddr:
            raise ValueError(f"{sub!r} does not fit into {one!r}")
        return sub.new(baseaddr=baseaddr, size=min(one.size, sub.size), is_sub=one.is_sub, words=sub.words)


def get_is_volatile(bus: Access | None, core: Access | None) -> bool:
    """Calc Volatile Flag based on Accesses."""
    if bus is not None:
        if bus.read and bus.read.data is not None:
            # Read operation on bus manipulates data
            return True
        if core is None:
            # Bus Access Only
            return False
    if core is not None:
        if core.read and core.read.data is not None:
            # Read operation on core side manipulates data
            return True
        if bus is not None and bus.write and core.write:
            # Two-Sides can manipulate
            return True
    return False


def get_is_const(bus: Access | None, core: Access | None) -> bool:
    """Calc Is Constant Flag based on Accesses."""
    if bus is not None:
        if bus.write is not None:
            return False
        if bus.read and bus.read.data is not None:
            return False
    if core is not None:
        if core.read and core.read.data is not None:
            return False
        if core.write is not None:
            return False
    return True


__dec_to_alpha = ("", *ascii_lowercase)


def name_alpha(num: int) -> str:
    """
    Convert number to a alpha digit.

    >>> name_alpha(0)
    'a'
    >>> name_alpha(25)
    'z'
    >>> name_alpha(26)
    'aa'
    >>> name_alpha(27)
    'ab'
    >>> name_alpha(26+25)
    'az'
    >>> name_alpha(26+26)
    'ba'
    >>> name_alpha(26+26+25)
    'bz'
    >>> name_alpha(1000)
    'alm'
    """
    num += 1

    if num < _ALPHA_PER_DIGIT:
        return __dec_to_alpha[num]

    result: list[str] = []
    while num:
        num, rem = divmod(num, _ALPHA_PER_DIGIT)
        result.insert(0, __dec_to_alpha[rem])
        if not rem:
            num -= 1
            result.insert(0, "z")
    return "".join(result)


def create_fill_word(addrspace, idx, offset, depth) -> Word:
    """Create Fill Word."""
    return Word(name=f"reserved{idx}", offset=offset, depth=depth, width=addrspace.width)


def create_fill_field(word, idx, offset, width) -> Field:
    """Create Fill Field."""
    return Field(name=f"reserved{idx}", type_=u.UintType(width), offset=offset)


class ReservedAddrspace(Addrspace):
    """A Reserved Address Space."""


class DefaultAddrspace(Addrspace):
    """Default Address Space."""

    name: str = u.Field(init=False, default="")


def resolve_field_value(field: Field, value: int | str) -> int:
    """Resolve Enumeration If necessary."""
    type_ = field.type_
    if value == "<RST>":
        return field.type_.default
    if not isinstance(type_, u.BaseEnumType) or not isinstance(value, str):
        return int_(value)[0]
    for item in type_.values():
        if item.value == value:
            return item.key
    # value is a string but not an enum value, so we need to check if
    # it is castable to int, might be an int, bit or hex string
    if isinstance(value, str):
        try:
            val, _ = int_(value)
            return val
        except ValueError:
            # drop down to our general, prettier error message below
            pass
    raise ValueError(f"Value {value!r} is not covered by {type_}")


def get_mask(word: Word, filter_=None) -> u.Hex:
    """
    Return mask for all fields of one word.

    >>> import ucdp_addr as ua
    >>> word = ua.Word(name='word', offset=0, width=32)
    >>> field = word.add_field('field0', u.UintType(3), ua.access.RO)
    >>> field = word.add_field('field1', u.UintType(6), ua.access.WO, align=4)
    >>> field = word.add_field('field2', u.UintType(3), ua.access.RW, align=4)

    >>> get_mask(word)
    Hex('0x000073F7')
    >>> get_mask(word, filter_=lambda field: field.bus.read)
    Hex('0x00007007')
    >>> get_mask(word, filter_=lambda field: field.bus.write)
    Hex('0x000073F0')
    >>> get_mask(word, filter_=lambda field: field.bus.read and field.bus.write)
    Hex('0x00007000')
    """
    fields = word.fields
    if filter_:
        fields = [field for field in fields if filter_(field)]
    mask = sum([field.slice.mask for field in fields])
    return u.Hex(mask, width=word.width)


def read_on_modify(word: Word, mask: int) -> bool:
    """
    Return `True` if a read is needed in read-modify-write operations on word `word` with `mask`.

    >>> import ucdp_addr as ua
    >>> word = ua.Word(name='word', offset=0, width=32)
    >>> field = word.add_field('field0', u.UintType(3), ua.access.RO)
    >>> field = word.add_field('field1', u.UintType(6), ua.access.WO, align=4)
    >>> field = word.add_field('field2', u.UintType(3), ua.access.RW, align=4)

    >>> read_on_modify(word, 0x00007000)
    False
    >>> read_on_modify(word, 0x00003000)
    True
    """
    rwmask = get_mask(word, lambda field: field.bus.read and field.bus.write)
    return bool(rwmask) and rwmask != mask



"""Data Information."""

from collections.abc import Iterator
from typing import TypeAlias

import ucdp as u

from .addrrange import AddrRange
from .data import Data, DataType, check_data, get_datatype

AddrData: TypeAlias = tuple[int, int]
"""Address-Data Pair."""


class DataInfo(u.Object):
    """
    Data Information.

    Single Data:

        >>> datainfo = DataInfo.create(5)
        >>> datainfo
        DataInfo(datatype=SINGLE, data=5)
        >>> str(datainfo)
        'SINGLE: 5'
        >>> tuple(datainfo.iter(AddrRange(width=64, depth=8)))
        ((Hex('0x0'), 5),)
        >>> tuple(datainfo.iter(AddrRange(baseaddr=0x1000, width=64, depth=8)))
        ((Hex('0x1000'), 5),)

    Burst Data:

        >>> datainfo = DataInfo.create((5, 6, 7))
        >>> datainfo
        DataInfo(datatype=BURST, data=(5, 6, 7))
        >>> str(datainfo)
        'BURST: (5, 6, 7)'
        >>> tuple(datainfo.iter(AddrRange(width=64, depth=8)))
        ((Hex('0x0'), 5), (Hex('0x8'), 6), (Hex('0x10'), 7))
        >>> tuple(datainfo.iter(AddrRange(baseaddr=0x1000, width=64, depth=8)))
        ((Hex('0x1000'), 5), (Hex('0x1008'), 6), (Hex('0x1010'), 7))

    Scattered Data:

        >>> datainfo = DataInfo.create(((16, 5), (28, 6), (60, 7)))
        >>> datainfo
        DataInfo(datatype=SCAT, data=((16, 5), (28, 6), (60, 7)))
        >>> str(datainfo)
        'SCAT: ((16, 5), (28, 6), (60, 7))'
        >>> tuple(datainfo.iter(AddrRange(width=64, depth=8)))
        ((Hex('0x10'), 5), (Hex('0x1C'), 6), (Hex('0x3C'), 7))
        >>> tuple(datainfo.iter(AddrRange(baseaddr=0x1000, width=64, depth=8)))
        ((Hex('0x1010'), 5), (Hex('0x101C'), 6), (Hex('0x103C'), 7))

    Errors:

        >>> DataInfo.create('a')
        Traceback (most recent call last):
          ...
        TypeError: a
        >>> DataInfo.create(('a', 'b'))
        Traceback (most recent call last):
          ...
        TypeError: ('a', 'b')
    """

    datatype: DataType
    data: Data

    def __str__(self) -> str:
        return f"{self.datatype.name}: {self.data}"

    @staticmethod
    def create(data: Data) -> "DataInfo":
        """Create :any:`DataInfo` for `data`."""
        datatype = get_datatype(data)
        return DataInfo(datatype=datatype, data=data)

    def iter(self, addrrange: AddrRange) -> Iterator[AddrData]:
        """Iteratate over single address value pairs according to access."""
        yield from _iter_data(self.datatype, self.data, addrrange)


def _iter_data(datatype: DataType, data: Data, addrrange: AddrRange) -> Iterator[AddrData]:
    baseaddr = addrrange.baseaddr
    check_data(data, addrrange.width)
    if datatype == DataType.SINGLE:
        yield baseaddr, data
    elif datatype == DataType.BURST:
        wordsize = addrrange.wordsize
        for idx, value in enumerate(data):  # type: ignore[arg-type]
            yield u.Hex(baseaddr + idx * wordsize), value
    else:
        addrmask = addrrange.addrmask
        for addr, value in data:  # type: ignore[union-attr]
            yield baseaddr + (addr & addrmask), value



"""Access Type Description."""

from enum import Enum
from typing import Any, TypeAlias

Data: TypeAlias = int | tuple[int, ...] | list[int] | tuple[tuple[int, int], ...] | list[tuple[int, int]]
"""Data."""


class DataType(Enum):
    """Trans Type."""

    SINGLE = 0
    BURST = 1
    SCAT = 2

    def __repr__(self):
        return self.name


def check_data(data: Data, width: int):
    """Check Data."""
    high = 2**width - 1
    for idx, value in enumerate(_unify_data(data)):
        if value > high or value < 0:
            raise ValueError(f"value {value} at index {idx} exceeds limits [0, {high}]")


def _unify_data(data: Data) -> list[int]:
    if isinstance(data, int):
        return [data]
    if isinstance(data, (tuple, list)):
        if any(isinstance(item, int) for item in data):
            return data  # type: ignore[return-value]
        if any(isinstance(item, tuple) and len(item) == 2 for item in data):  # noqa: PLR2004
            return [value for _, value in data]
    raise TypeError(data)


def get_size(data: Any, wordsize: int) -> int:
    """Determine Maximum Addressed Size of Data."""
    if isinstance(data, int):
        return wordsize
    if isinstance(data, (tuple, list)):
        if any(isinstance(item, int) for item in data):
            return wordsize * len(data)
        if any(isinstance(item, tuple) and len(item) == 2 for item in data):  # noqa: PLR2004
            addrs = [addr for addr, _ in data]  # type:ignore[union-attr]
            return max(addrs) + wordsize
    raise TypeError(data)


def get_datatype(data: Any) -> DataType:
    """Data Type."""
    if isinstance(data, int):
        return DataType.SINGLE
    if isinstance(data, (tuple, list)):
        if any(isinstance(item, int) for item in data):
            return DataType.BURST
        if any(isinstance(item, tuple) and len(item) == 2 for item in data):  # noqa: PLR2004
            return DataType.SCAT
    raise TypeError(data)
